from simses.technology.hydrogen.control.thermal.thermal_controller import ThermalController


class NoThermalController(ThermalController):

    def __init__(self):
        super().__init__()

    def calculate(self, stack_temperature: float, heat_stack: float, timestep: float, current_dens: float) -> None:
        pass

    def get_heat_control_on(self) -> bool:
        return True

    def get_h2o_flow(self) -> float:
        return 0.0

    def get_delta_water_temp_in(self) -> float:
        return 25.0



import math

from simses.technology.lithium_ion.cell.format.abstract import CellFormat


class RoundCell(CellFormat):
    
    def __init__(self, diameter: float, length: float):
        super(RoundCell, self).__init__()
        self.__volume = math.pi * diameter ** 2.0 * length / 4.0 * 10 ** (-9)  # m3
        self.__surface_area = (2 * math.pi * diameter / 2.0 * length + 2 * math.pi * (diameter / 2.0) ** 2.0) * 10 ** (-6)  # m2

    def get_volume(self) -> float:
        return self.__volume

    def get_surface_area(self) -> float:
        return self.__surface_area



from abc import ABC, abstractmethod


class TechnicalProfile(ABC):
    """
    Profile for additional data, e.g. frequency.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def next(self, time: float) -> float:
        """
        provides the data for a technical profile (SoC or frequency)

        Parameters
        ----------

        Returns
        -------
        float
            next data in a specific profile
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all open resources in market profile

        Parameters
        ----------

        Returns
        -------

        """
        pass



from calendar import isleap
from datetime import datetime

import pytz


def get_sum_for(data) -> float:
    return data[:].sum()


def get_mean_for(data) -> float:
    return data[:].mean()


def get_min_for(data) -> float:
    return data[:].min()


def get_max_for(data) -> float:
    return data[:].max()


def get_positive_values_from(data):
    _data = data[:].copy()
    _data[_data < 0] = 0
    return _data


def get_negative_values_from(data):
    _data = data[:].copy()
    _data[_data > 0] = 0
    return _data


def get_fractional_years(start_timestamp, end_timestamp):
    start_date = datetime.fromtimestamp(start_timestamp, tz=pytz.UTC)
    end_date = datetime.fromtimestamp(end_timestamp, tz=pytz.UTC)
    diffyears = end_date.year - start_date.year
    difference = end_date - start_date.replace(end_date.year)
    days_in_year = isleap(end_date.year) and 366 or 365
    total_fractional_years = diffyears + (difference.days + difference.seconds / 86400.0) / days_in_year
    return total_fractional_years



import random

import numpy as np
import pandas as pd


class MaxPeakShavingLimit:
    """
    Determines minimum peak shaving limit that can be served for a given load profile based on a energy storage
    bucket model.
    """

    def __init__(self, energy: float, power: float, efficiency: float, time_step: float, random_profile: bool, file_path: str = '', scaling_factor: float = 1):
        """

        Parameters
        ----------
        energy : energy in Wh
        power : max power in W on the AC side
        efficiency : efficiency in p.u.
        time_step : time step in s
        random_profile: choose to generate a random profile
        file_path: path for power file
        """
        self.__energy = energy
        self.ac__power_rating = power
        self.__efficiency = efficiency
        self.__time_step = time_step
        if random_profile:
            self.__profile = np.asarray([5e6 + random.randint(-1e6, 1e6) for i in range(35040)])
        else:
            self.__profile = self.__read_full_profile(file_path, time_step, scaling_factor)
        self.__ws_to_wh = 1 / 3600

    def calc_max_peak(self, max_runs: float) -> float:
        max_limit = max(self.__profile)
        min_limit = max(self.__profile) - self.ac__power_rating
        max_limit_valid = self.__check_for_valid_limit(max_limit)
        min_limit_valid = self.__check_for_valid_limit(min_limit)

        # if rated power is a valid limit, use the rated power
        if min_limit_valid:
            print('Minimum Peak Shaving Limit is equal to Max Load Power - Nom. Storage Power: \t' + str(round(min_limit / 1e3, 2)) + ' kW')
            return min_limit
        if not max_limit_valid:
            return 0
        print('Limits after run ' + str(0) + '\t\tMin: ' + '{:.2f}'.format(min_limit / 1e3, 4)
              + ' kW\t\tMax: ' + '{:.2f}'.format(max_limit / 1e3, 4) + ' kW')

        run = 0
        while run < max_runs:
            mid_power = (max_limit + min_limit) / 2
            mid_valid = self.__check_for_valid_limit(mid_power)
            if mid_valid:
                max_limit = mid_power
            else:
                min_limit = mid_power
            print('Limits after run ' + str(run) + '\t\tMin: ' + '{:.2f}'.format(min_limit / 1e3, 4)
                  + ' kW\t\tMax: ' + '{:.2f}'.format(max_limit / 1e3, 4) + ' kW')
            run += 1

        print('\nRESULTS:')
        print('Maximum Peak Shaving Limit: \t' + str(round((max_limit + min_limit) / 2 / 1e3, 2)) + ' kW')
        abs_error = (max_limit - min_limit) / 2
        print('Absolute Error: \t\t\t\t' + str(round(abs_error, 2)) + ' W')
        rel_error = ((max_limit - min_limit) / 2) / ((max_limit + min_limit) / 2) * 100
        print('Relative Error: \t\t\t\t' + str(round(rel_error, 4)) + ' %')
        return min_limit

    def __check_for_valid_limit(self, limit: float) -> bool:
        soc = 1
        net_load_series = (self.__profile - limit)

        for i in range(len(net_load_series)):
            if net_load_series[i] > self.ac__power_rating:
                dc_storage_power = self.ac__power_rating / self.__efficiency
            elif net_load_series[i] < -1 * self.ac__power_rating:
                dc_storage_power = -1 * self.ac__power_rating * self.__efficiency
            else:
                if net_load_series[i] >= 0:
                    dc_storage_power = net_load_series[i] / self.__efficiency
                else:
                    dc_storage_power = net_load_series[i] * self.__efficiency
            # Positive dc_storage_power while charging
            soc += -1 * dc_storage_power * self.__time_step * self.__ws_to_wh / self.__energy
            # If SOC hits 0, the peak cannot be shaved
            if soc < 0:
                return False
            if soc > 1:
                soc = 1
        return True

    def __read_full_profile(self, file_path, sample_time, scaling_factor):
        load_profile_in = pd.read_csv(file_path, delimiter='\n', header=None,)
        print('Done Reading Profile.')

        load_profile = load_profile_in.to_numpy()
        load_profile = load_profile.transpose()
        load_profile = load_profile[0, :]
        original_load_profile_length = load_profile.size
        original_seconds_per_timestep = (8760 / original_load_profile_length) * 3600
        original_timesteps = np.arange(1, 8761, original_seconds_per_timestep/3600)
        np.place(original_timesteps, original_timesteps > 8760, 8760)
        to_interpolate_time_steps = np.arange(1, 8761, sample_time / 3600)
        np.place(to_interpolate_time_steps, to_interpolate_time_steps > 8760, 8760)
        interpolated_load_profile = np.interp(to_interpolate_time_steps, original_timesteps, load_profile) * scaling_factor
        return interpolated_load_profile


if __name__ == '__main__':
    scaling_factor = 1
    path = r"C:\Users\Stefan\Desktop\peakshaving_profil3.csv"
    random_profile = False
    limitfinder = MaxPeakShavingLimit(1370e3, 1e6, 0.9, 600, random_profile, path, scaling_factor)
    limitfinder.calc_max_peak(20)



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.constants import Hydrogen
from simses.technology.hydrogen.control.thermal.thermal_controller import ThermalController


class VarFlowThermalController(ThermalController):
    """
    This controller controls the temperature of the EL-stack by varing the input temperaure of the feeding water and
    in the second step adapt the mass flow of the water in order to reach the desired temperature
    """

    def __init__(self, config: ElectrolyzerConfig, min_water_flow_stack: float, heat_capacity: float):
        super().__init__()
        self.__stack_temp_desire = config.desire_temperature # °C
        self.__slope_faktor = - 1 / 5
        self.__delta_water_temperature = 5  # K
        self.__max_cooling_rate = 3  # K/s; assumption needs to be that high, so that there is no thermal runaway
        self.__temperature_h2o_in = 0  # °C
        self.__h2o_flow = 0  # mol/s#
        self.__min_water_flow_stack: float = min_water_flow_stack
        self.__heat_capacity: float = heat_capacity

    def calculate(self, stack_temperature: float, heat_stack: float, timestep: float, current_dens: float) -> None:
        temp_diff = self.__stack_temp_desire - stack_temperature
        # calculation of water temperature
        control_faktor = self.__calculate_control_faktor_temperature(temp_diff)
        self.__temperature_h2o_in = stack_temperature - control_faktor * self.__delta_water_temperature
        # calculation of water flow
        if stack_temperature <= self.__stack_temp_desire:
            self.__h2o_flow = self.__min_water_flow_stack
        else:
            ideal_cooling_rate = temp_diff / timestep
            if ideal_cooling_rate < self.__max_cooling_rate:
                cooling_rate = ideal_cooling_rate
            else:
                cooling_rate = self.__max_cooling_rate
            self.__h2o_flow = self.__heat_capacity * cooling_rate / (Hydrogen.HEAT_CAPACITY_WATER *
                                                                     Hydrogen.MOLAR_MASS_WATER * temp_diff)

    def get_delta_water_temp_in(self) -> float:
        return self.__temperature_h2o_in

    def get_h2o_flow(self) -> float:
        return self.__h2o_flow

    def __calculate_control_faktor_temperature(self, temp_diff) -> float:
        if abs(temp_diff) < 5:
            return self.__slope_faktor * temp_diff
        else:
            return 1.0



import statistics
from scipy.integrate import solve_ivp
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.system import SystemState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter
from simses.system.power_electronics.dcdc_converter.abstract_dcdc_converter import DcDcConverter
from simses.system.storage_system_dc import StorageSystemDC
from simses.system.thermal.ambient.ambient_thermal_model import AmbientThermalModel
from simses.system.thermal.model.system_thermal_model import SystemThermalModel
from simses.technology.storage import StorageTechnology


class ZeroDRoomThermalModel(SystemThermalModel):

    def __init__(self, ambient_thermal_model: AmbientThermalModel, general_config: GeneralSimulationConfig,
                 dc_systems: [StorageSystemDC], acdc_converter: AcDcConverter):

        super().__init__()

        # Components
        self.__acdc_converter = acdc_converter
        self.__dc_systems: [StorageSystemDC] = dc_systems
        self.__storage_technologies: [StorageTechnology] = list()
        self.__dc_dc_converters: [DcDcConverter] = list()
        for dc_system in self.__dc_systems:  # Unpack storage technologies and DC/DC converters from StorageSystemDC
            self.__storage_technologies.append(dc_system.get_storage_technology())
            self.__dc_dc_converters.append(dc_system.get_dc_dc_converter())

        # Simulation time parameters
        self.__sample_time: float = general_config.timestep
        self.__ts_adapted = 0

        # Models
        self.__ambient_thermal_model: AmbientThermalModel = ambient_thermal_model

        # Auto-adjust thermal model calculation timestep
        if general_config.end - general_config.start >= 86400:  # Choose a larger calculation_time_step for faster performance
            self.__calculation_time_step = 300
        else:
            self.__calculation_time_step = 60
        if self.__calculation_time_step >= self.__sample_time:
            self.__calculation_time_step = int(self.__sample_time)

        if self.__sample_time % self.__calculation_time_step != 0:
            raise Exception(self.__name__ +
                            ': For simulations longer than a day, please set a simulation timestep '
                            'in multiples of 5 minutes, for shorter simulations, use a timestep in multiples of 1 minute.')

        # -- Evaluation Time --
        # self.__t_eval_step is the time step after which the equation gets evaluated, it has impact on the I-controller
        # and the plotting (it results in the graphs getting more detailed).
        # Set t_eval-step to self.__calculation_time_step if plots are not needed because the impact on the I-Controller
        # is negligible.
        self.__t_eval_step: int = self.__calculation_time_step

        # Initialze temperatures of BESS components with the HVAC set-point temperature
        self.__internal_air_temperature: float = self.__ambient_thermal_model.get_initial_temperature()  # in K
        self.__storage_technology_temperatures = len(self.__storage_technologies) \
                                                 * [self.__ambient_thermal_model.get_initial_temperature()]  # K
        self.__converter_temperature_ac_dc = self.__ambient_thermal_model.get_initial_temperature()  # in K

        # Initialize internal air parameters
        # Model with p & V constant, i.e. if T rises, mass must decrease.
        # Quantities with reference to ideal gas equation
        self.__individual_gas_constant = self.universal_gas_constant / self.molecular_weight_air  # J/kgK
        self.__air_density = self.air_pressure / (self.__individual_gas_constant * 298.15)  # kg/m3

        # StorageTechnology thermal paramaters
        self.__surface_area_storage_technology = list()
        self.__mass_storage_technology = list()
        self.__specific_heat_capacity_storage_technology = list()
        self.__convection_coefficient_storage_technology_air = list()
        for storage_technology in self.__storage_technologies:
            self.__surface_area_storage_technology.append(storage_technology.surface_area)  # in m2
            self.__mass_storage_technology.append(storage_technology.mass)  # in kg
            self.__specific_heat_capacity_storage_technology.append(storage_technology.specific_heat)  # in J/kgK
            self.__convection_coefficient_storage_technology_air.append(storage_technology.convection_coefficient)  # in W/m2K

        # DC/DC Converter thermal parameters
        # Note: in this model the DC-DC Converter is treated as an part of the battery. That means battery and DC-DC
        # converter are seen as one thermal component with the same specific heat capacity and convection coefficient.
        # For the simulation/calculation the mass, surface area and thermal losses of the DC-DC converter are
        # added to the respective battery values.
        self.__surface_area_converter_dc_dc = list()
        self.__mass_converter_dc_dc = list()
        for dc_dc_converter in self.__dc_dc_converters:
            self.__surface_area_converter_dc_dc.append(dc_dc_converter.surface_area)  # in m2
            self.__mass_converter_dc_dc.append(dc_dc_converter.mass)  # in kg

        # AC/DC Converter thermal parameters
        self.__surface_area_converter_ac_dc = self.__acdc_converter.surface_area  # in m2
        self.__mass_converter_ac_dc = self.__acdc_converter.mass  # in kg
        self.__specific_heat_capacity_converter_ac_dc = statistics.mean(self.__specific_heat_capacity_storage_technology)  # in J/kgK
        self.__convection_coefficient_converter_ac_dc = statistics.mean(self.__convection_coefficient_storage_technology_air)  # in W/m2K

        self.__calculate_thermal_resistances()
        self.__calculate_thermal_capacities()

        self.__ambient_air_temperature = self.__ambient_thermal_model.get_initial_temperature()
        self.__solar_irradiation_thermal_load = 0
        self.__hvac_thermal_power = 0

    def __calculate_thermal_resistances(self) -> None:
        """
        calculates and sets the values of all thermal resistances in the model (in K/W)
        :return: None
        """
        # all units in K/W
        # calculates thermal convection resistance between the surrounding air (sa) and  the margin of Layer 3 (l3)

        # calculates thermal convection resistance  between the storage technologies and the internal air
        self.__storage_technology_ia_thermal_resistance = list()
        for storage_technology in self.__storage_technologies:
            index = self.__storage_technologies.index(storage_technology)
            self.__storage_technology_ia_thermal_resistance.append(
                1 / (self.__convection_coefficient_storage_technology_air[index] *
                     (self.__surface_area_storage_technology[index] + self.__surface_area_converter_dc_dc[index])))

        # calculates thermal convection resistance  between the converter and the inner air (ia)
        self.__converter_ia_thermal_resistance = 1 / (self.__convection_coefficient_converter_ac_dc *
                                                      self.__surface_area_converter_ac_dc)

    def __calculate_thermal_capacities(self) -> None:
        """
        calculates and sets the values of all thermal capacities in the model (in J/K)
        :return: None
        """
        # all units in J/K
        self.__storage_technology_thermal_capacity = list()
        for storage_technology in self.__storage_technologies:
            index = self.__storage_technologies.index(storage_technology)
            self.__storage_technology_thermal_capacity.append(
                (self.__mass_storage_technology[index] + self.__mass_converter_dc_dc[index]) * \
                self.__specific_heat_capacity_storage_technology[index])

        self.__converter_thermal_capacity = self.__mass_converter_ac_dc * self.__specific_heat_capacity_converter_ac_dc

    def update_air_parameters(self) -> None:
        """
        updates values of mass and air density stored within the HVAC object
        :return: None
        """
        pass

    def calculate_temperature(self, time, state: SystemState, storage_system_dc_states: [SystemState]) -> None:
        """
        primary method of the room thermal model which calculates and sets temperatures of all components
        :param time: timestamp of current timestep as float
        :param state: state of the StorageSystemAC as SystemState
        :param storage_system_dc_states: states of the StorageSystemDC objects within StorageSystemAC as [SystemState]
        :return: None
        """
        ambient_air_temperature = self.__ambient_thermal_model.get_temperature(time - self.__ts_adapted)
        self.__ambient_air_temperature = ambient_air_temperature
        self.__internal_air_temperature = ambient_air_temperature  # Housing is not present
        self.__air_density = self.air_pressure / (self.__individual_gas_constant * state.temperature)

        calculated_time = 0
        radiation_power = 0
        self.__solar_irradiation_thermal_load = radiation_power
        hvac_electric_consumption = []
        thermal_power_hvac = []

        while calculated_time < self.__sample_time:
            calculated_time += self.__calculation_time_step
            thermal_power = 0.0
            thermal_power_hvac.append(thermal_power)

            def equation_rhs(t, variable_array):
                """
                specifies the set of simultaneous differential equations to be solved
                """
                # variable_array = [inner_air_temperature,
                # storage_technology_temperatures(list, len = # storage technologies), converter_temperature]
                # Temperature variables: inner_air_temperature, storage_technology_temperatures, converter_temperature
                # independent variable: time

                number_storage_technologies = len(self.__storage_technologies)
                d_by_dt_storage_technology_temperature = list()
                heat_flow_storage_technology_ia = list()
                for storage_technology in self.__storage_technologies:
                    index = self.__storage_technologies.index(storage_technology)
                    storage_system_dc_state = storage_system_dc_states[index]
                    heat_flow_storage_technology_ia.append((variable_array[index + 1] - variable_array[0]) /
                                                           self.__storage_technology_ia_thermal_resistance[index])
                    # Differential equation for change in storage technology temperature
                    d_by_dt_storage_technology_temperature.append(
                        ((storage_system_dc_state.storage_power_loss + storage_system_dc_state.dc_power_loss) -
                         (variable_array[index + 1] - variable_array[0]) /
                         self.__storage_technology_ia_thermal_resistance[index]) / \
                        self.__storage_technology_thermal_capacity[index])

                # Differential equation for change in inner air temperature
                d_by_dt_inner_air_temperature = 0.0

                # Differential equation for change in converter temperature
                d_by_dt_converter_temperature_ac_dc = (state.pe_losses - (
                        (variable_array[number_storage_technologies + 1] - variable_array[0])
                        / self.__converter_ia_thermal_resistance)) / self.__converter_thermal_capacity

                equation_rhs_array = [d_by_dt_inner_air_temperature] + d_by_dt_storage_technology_temperature + \
                                     [d_by_dt_converter_temperature_ac_dc]
                return equation_rhs_array

            # time_interval is an array of times at which the equation get evaluated
            time_interval = [i for i in range(self.__t_eval_step, self.__calculation_time_step + self.__t_eval_step,
                                              self.__t_eval_step)]
            storage_technology_temperatures = self.__storage_technology_temperatures
            temperature_variable_array = [self.__internal_air_temperature] + storage_technology_temperatures + \
                                         [self.__converter_temperature_ac_dc]

            sol = solve_ivp(equation_rhs, (0, self.__calculation_time_step),
                            temperature_variable_array,
                            method='BDF', t_eval=time_interval)

            temperature_series = sol.y

            # setting temperatures for the next iteration of the while loop
            self.__internal_air_temperature = temperature_series[0, -1]
            number_storage_technologies = len(self.__storage_technologies)
            i = 0
            while i < len(self.__storage_technologies):
                self.__storage_technology_temperatures[i] = temperature_series[i + 1, -1]
                i += 1
            self.__converter_temperature_ac_dc = temperature_series[number_storage_technologies + 1, -1]

            hvac_electric_consumption.append(0.0)

        # setting storage technology temperature for SIMSES simulation and plotting
        for storage_technology in self.__storage_technologies:
            index = self.__storage_technologies.index(storage_technology)
            storage_technology.state.temperature = self.__storage_technology_temperatures[index]

        # Set HVAC electric consumption over sample time
        self.__hvac_thermal_power = 0.0

    def get_auxiliaries(self) -> [Auxiliary]:
        """
        :returns instances of the Auxiliary class (HVAC system in this case)
        :return: __heating_cooling as [Auxiliary]
        """
        return list()

    def get_temperature(self) -> float:
        """
        returns internal air temperature in the container in K
        :return: __internal_air_temperature as float
        """
        return self.__internal_air_temperature

    def get_ambient_temperature(self) -> float:
        """
        returns the ambient temperature at the location in K
        :return: __ambient_air_temperature as float
        """
        return self.__ambient_air_temperature

    def get_solar_irradiation_thermal_load(self) -> float:
        """
        returns the thermal load due to solar irradiation in W
        :return: __solar_irradiation_thermal_load as float
        """
        return self.__solar_irradiation_thermal_load

    def get_hvac_thermal_power(self) -> float:
        """
        returns the HVAC thermal power in W
        :return: __hvac_thermal_power as float
        """
        return self.__hvac_thermal_power

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        self.__ts_adapted = ts_adapted
        self.__ambient_thermal_model: AmbientThermalModel = self.__ambient_thermal_model.create_instance()

    def close(self) -> None:
        """
        closes specified open resources
        """
        self.__ambient_thermal_model.close()

    def get_ol_temperature(self) -> float:
        """
        Not applicable to this class
        """
        pass

    def get_il_temperature(self) -> float:
        """
        Not applicable to this class
        """
        pass



from simses.technology.lithium_ion.cell.format.round import RoundCell


class RoundCell18650(RoundCell):

    __DIAMETER: float = 18  # mm
    __LENGTH: float = 65  # mm

    def __init__(self):
        super(RoundCell18650, self).__init__(diameter=self.__DIAMETER, length=self.__LENGTH)



from configparser import ConfigParser

from simses.commons.data.data_handler import DataHandler
from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.commons.state.technology.hydrogen import HydrogenState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.technology.hydrogen.control.management import HydrogenManagementSystem
from simses.technology.hydrogen.electrolyzer.system import Electrolyzer
from simses.technology.hydrogen.factory import HydrogenFactory
from simses.technology.hydrogen.fuel_cell.system import FuelCell
from simses.technology.hydrogen.hydrogen_storage.hydrogen_storage import HydrogenStorage
from simses.technology.storage import StorageTechnology


class Hydrogen(StorageTechnology):

    """
    Hydrogen is the top level class for coordinating all technologies related to hydrogen, i.e. electrolyzer, fuel cells
    and storage tank. It provides a management system for taking limitations into consideration. The hydrogen class
    requires always an electrolyzer, a fuel cell and a storage. In cases this circle is unwanted, dummy classes
    exists doing nothing, e.g. NoElectrolyzer or NoFuelCell.
    """

    def __init__(self, data_export: DataHandler, fuel_cell: str, fuel_cell_maximal_power: float, electrolyzer: str,
                 electrolyzer_maximal_power: float, storage: str, capacity: float, max_pressure: float,
                 temperature: float, system_id: int, storage_id: int, config: ConfigParser):
        super().__init__()
        self.__log = Logger(type(self).__name__)
        self.__electrolyzer: Electrolyzer = Electrolyzer(system_id, storage_id, electrolyzer,
                                                         electrolyzer_maximal_power, temperature,
                                                         config, data_export)
        self.__fuel_cell: FuelCell = FuelCell(system_id, storage_id, fuel_cell, fuel_cell_maximal_power,
                                              temperature, config, data_export)
        factory: HydrogenFactory = HydrogenFactory(config)
        # TODO Get rid of hard coded crap (MM)
        if electrolyzer == "AlkalineElectrolyzer":
            electrolyzer_minimal_power = electrolyzer_maximal_power * 0.2
        else:
            electrolyzer_minimal_power = 0
        self.__management_system: HydrogenManagementSystem = factory.create_hydrogen_management_system(electrolyzer_minimal_power,
                                                                electrolyzer_maximal_power, fuel_cell_maximal_power)
        self.__hydrogen_storage: HydrogenStorage = factory.create_hydrogen_storage(storage, capacity, max_pressure)
        self.__hydrogen_state: HydrogenState = factory.create_hydrogen_state_from(system_id, storage_id)
        self.__hydrogen_state.max_charge_power = electrolyzer_maximal_power
        self.__hydrogen_state.max_discharge_power = fuel_cell_maximal_power
        factory.close()
        self.__update_state(self.__hydrogen_state.time)

    def distribute_and_run(self, time: float, current: float, voltage: float):
        hs: HydrogenState = self.__hydrogen_state
        hs.power = current * voltage
        self.__management_system.update(time, hs)
        self.__electrolyzer.update(time, hs.power)
        self.__fuel_cell.update(time, hs.power)
        self.__hydrogen_storage.update_from(time - hs.time, self.__electrolyzer.state, self.__fuel_cell.state)
        self.__update_state(time)
        self.__log.info('Done with update')

    def __update_state(self, time: float) -> None:
        hydrogen_state: HydrogenState = self.__hydrogen_state
        electrolyzer_state: ElectrolyzerState = self.__electrolyzer.state
        fuel_cell_state: FuelCellState = self.__fuel_cell.state
        if hydrogen_state.is_charge:
            hydrogen_state.current = electrolyzer_state.current
            hydrogen_state.voltage = electrolyzer_state.voltage
            hydrogen_state.fulfillment *= electrolyzer_state.fulfillment
            hydrogen_state.power_loss = electrolyzer_state.power_loss
            hydrogen_state.temperature = electrolyzer_state.temperature
        else:
            hydrogen_state.current = fuel_cell_state.current
            hydrogen_state.voltage = fuel_cell_state.voltage
            hydrogen_state.fulfillment *= fuel_cell_state.fulfillment
            hydrogen_state.power_loss = fuel_cell_state.power_loss
            hydrogen_state.temperature = fuel_cell_state.temperature
        hydrogen_state.soc = self.__hydrogen_storage.get_soc()
        hydrogen_state.capacity = self.__hydrogen_storage.get_capacity()
        hydrogen_state.soh = electrolyzer_state.soh * fuel_cell_state.soh
        hydrogen_state.time = time

    @property
    def volume(self) -> float:
        return 0

    @property
    def mass(self) -> float:
        raise NotImplementedError('Mass is not implemented in hydrogen yet')

    @property
    def surface_area(self) -> float:
        raise NotImplementedError('Surface area is not implemented in hydrogen yet')

    @property
    def specific_heat(self) -> float:
        raise NotImplementedError('Specific heat is not implemented in hydrogen yet')

    @property
    def convection_coefficient(self) -> float:
        raise NotImplementedError('Convection coefficient is not implemented in hydrogen yet')

    def wait(self):
        pass

    def get_auxiliaries(self) -> [Auxiliary]:
        aux: [Auxiliary] = list()
        aux.extend(self.__electrolyzer.get_auxiliaries())
        aux.extend(self.__fuel_cell.get_auxiliaries())
        aux.extend(self.__hydrogen_storage.get_auxiliaries())
        return aux

    @property
    def state(self):
        return self.__hydrogen_state

    def get_system_parameters(self) -> dict:
        return dict()

    def close(self):
        self.__management_system.close()
        self.__electrolyzer.close()
        self.__fuel_cell.close()
        self.__hydrogen_storage.close()
        self.__log.close()



from configparser import ConfigParser

from simses.commons.config.simulation.simulation_config import SimulationConfig


class RedoxFlowConfig(SimulationConfig):
    """
    Redox Flow specific configs
    """

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)
        self.__section: str = 'REDOX_FLOW'

    @property
    def soc(self) -> float:
        """Returns start soc for rfb from __analysis_config file_name"""
        return float(self.get_property(self.__section, 'START_SOC'))

    @property
    def min_soc(self) -> float:
        """Returns min soc for rfb from __analysis_config file_name"""
        return float(self.get_property(self.__section, 'MIN_SOC'))

    @property
    def max_soc(self) -> float:
        """Returns max soc for rfb from __analysis_config file_name"""
        return float(self.get_property(self.__section, 'MAX_SOC'))

    @property
    def exact_size(self) -> bool:
        """Returns selection for exact sizing True/False"""
        return self.get_property(self.__section, 'EXACT_SIZE') in ['True']



import pandas

from simses.analysis.data.abstract_data import Data
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.energy_management import EnergyManagementState


class EnergyManagementData(Data):
    """
    Provides time series data from EnergyManamgentState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)

    @property
    def id(self) -> str:
        return '0.0'

    @property
    def time(self):
        return self._get_data(EnergyManagementState.TIME)

    @property
    def power(self):
        pass

    @property
    def dc_power(self):
        pass

    @property
    def energy_difference(self):
        pass

    @property
    def soc(self):
        pass

    @property
    def capacity(self):
        pass

    @property
    def state_of_health(self):
        pass

    @property
    def storage_fulfillment(self):
        pass

    @property
    def load_power(self):
        return self._get_data(EnergyManagementState.LOAD_POWER)

    @property
    def pv_power(self):
        return self._get_data(EnergyManagementState.PV_POWER)

    @property
    def fcr_max_power(self):
        return self._get_data(EnergyManagementState.FCR_MAX_POWER)

    @property
    def idm_power(self):
        return self._get_data(EnergyManagementState.IDM_POWER)

    @property
    def peakshaving_limit(self):
        return self._get_data(EnergyManagementState.PEAKSHAVING_LIMIT)

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, EnergyManagementState, EnergyManagementState.TIME)
        res: [EnergyManagementData] = list()
        for data in system_data:
            res.append(EnergyManagementData(config, data))
        return res



from simses.commons.state.abstract_state import State


class SystemState(State):
    """
    Current physical state of the system with the main electrical parameters.
    """

    SYSTEM_AC_ID = 'StorageSystemAC'
    SYSTEM_DC_ID = 'StorageSystemDC'
    AC_POWER = 'AC power in W'
    PE_LOSSES = 'PE losses in W'
    AC_FULFILLMENT = 'AC Fulfillment in p.u.'
    AC_POWER_DELIVERED = 'AC power delivered in W'
    # DC_VOLTAGE_INPUT = 'DC voltage input in V'
    # DC_VOLTAGE = 'DC voltage in V'
    DC_CURRENT = 'DC current in A'
    AUX_LOSSES = 'Aux losses in W'
    DC_POWER_INTERMEDIATE_CIRCUIT = 'DC power of intermediate circuit in W'
    DC_POWER_STORAGE = 'DC power of storage in W'
    DC_POWER_LOSS = 'DC power loss in W'
    DC_POWER_ADDITIONAL = 'DC power additional in W'
    DC_VOLTAGE_CIRCUIT = 'DC voltage of intermediate circuit in V'
    STORAGE_POWER_LOSS = 'DC power loss of storage technology in W'
    SOC = 'SOC in p.u.'
    SOH = 'State of Health in p.u.'
    CAPACITY = 'Capacity in Wh'

    MAX_CHARGE_POWER = 'Maximum charging power in W'
    MAX_DISCHARGE_POWER = 'Maximum discharging power in W'

    TEMPERATURE = 'Internal air temperature in K'
    AMBIENT_TEMPERATURE = 'Ambient temperature in K'
    HVAC_THERMAL_POWER = 'HVAC thermal power in W'
    SOLAR_THERMAL_LOAD = 'Solar irradiation thermal load in W'

    # ACDC_TEMPERATURE = 'ACDC converter temperature in K'

    OL_TEMPERATURE = 'Outer layer temperature in K'
    IL_TEMPERATURE = 'Inner layer temperature in K'

    def __init__(self, system_id: int, storage_id: int):
        super().__init__()
        self._initialize()
        self.set(self.SYSTEM_AC_ID, system_id)
        self.set(self.SYSTEM_DC_ID, storage_id)

    @property
    def ol_temperature(self) -> float:
        return self.get(self.OL_TEMPERATURE)

    @ol_temperature.setter
    def ol_temperature(self, value: float) -> None:
        self.set(self.OL_TEMPERATURE, value)

    @property
    def il_temperature(self) -> float:
        return self.get(self.IL_TEMPERATURE)

    @il_temperature.setter
    def il_temperature(self, value: float) -> None:
        self.set(self.IL_TEMPERATURE, value)

    @property
    def ac_power(self) -> float:
        return self.get(self.AC_POWER)

    @ac_power.setter
    def ac_power(self, value: float) -> None:
        self.set(self.AC_POWER, value)

    @property
    def pe_losses(self) -> float:
        return self.get(self.PE_LOSSES)

    @pe_losses.setter
    def pe_losses(self, value: float) -> None:
        self.set(self.PE_LOSSES, value)

    @property
    def ac_fulfillment(self) -> float:
        return self.get(self.AC_FULFILLMENT)

    @ac_fulfillment.setter
    def ac_fulfillment(self, value: float) -> None:
        """
        Ratio of provided AC power to requested AC power from the power distributor for each timestep in p.u.

        Returns
        -------

        """
        self.set(self.AC_FULFILLMENT, value)

    @property
    def ac_power_delivered(self) -> float:
        return self.get(self.AC_POWER_DELIVERED)

    @ac_power_delivered.setter
    def ac_power_delivered(self, value: float) -> None:
        self.set(self.AC_POWER_DELIVERED, value)

    # @property
    # def dc_voltage_input(self) -> float:
    #     return self.get(self.DC_VOLTAGE_INPUT)
    #
    # @dc_voltage_input.setter
    # def dc_voltage_input(self, value: float) -> None:
    #     self.set(self.DC_VOLTAGE_INPUT, value)

    # @property
    # def voltage(self) -> float:
    #     return self.get(self.DC_VOLTAGE)
    #
    # @voltage.setter
    # def voltage(self, value: float) -> None:
    #     self.set(self.DC_VOLTAGE, value)

    @property
    def dc_current(self) -> float:
        return self.get(self.DC_CURRENT)

    @dc_current.setter
    def dc_current(self, value: float) -> None:
        self.set(self.DC_CURRENT, value)

    @property
    def aux_losses(self) -> float:
        return self.get(self.AUX_LOSSES)

    @aux_losses.setter
    def aux_losses(self, value: float) -> None:
        self.set(self.AUX_LOSSES, value)

    @property
    def dc_power_intermediate_circuit(self) -> float:
        return self.get(self.DC_POWER_INTERMEDIATE_CIRCUIT)

    @dc_power_intermediate_circuit.setter
    def dc_power_intermediate_circuit(self, value: float) -> None:
        self.set(self.DC_POWER_INTERMEDIATE_CIRCUIT, value)

    @property
    def dc_power_loss(self) -> float:
        return self.get(self.DC_POWER_LOSS)

    @dc_power_loss.setter
    def dc_power_loss(self, value: float) -> None:
        self.set(self.DC_POWER_LOSS, value)

    @property
    def storage_power_loss(self) -> float:
        return self.get(self.STORAGE_POWER_LOSS)

    @storage_power_loss.setter
    def storage_power_loss(self, value: float) -> None:
        self.set(self.STORAGE_POWER_LOSS, value)

    @property
    def dc_power_additional(self) -> float:
        return self.get(self.DC_POWER_ADDITIONAL)

    @dc_power_additional.setter
    def dc_power_additional(self, value: float) -> None:
        self.set(self.DC_POWER_ADDITIONAL, value)

    @property
    def dc_power_storage(self) -> float:
        return self.get(self.DC_POWER_STORAGE)

    @dc_power_storage.setter
    def dc_power_storage(self, value: float) -> None:
        self.set(self.DC_POWER_STORAGE, value)

    @property
    def dc_circuit_voltage(self) -> float:
        return self.get(self.DC_VOLTAGE_CIRCUIT)

    @dc_circuit_voltage.setter
    def dc_circuit_voltage(self, value: float) -> None:
        self.set(self.DC_VOLTAGE_CIRCUIT, value)

    @property
    def soc(self) -> float:
        return self.get(self.SOC)

    @soc.setter
    def soc(self, value: float) -> None:
        self.set(self.SOC, value)

    @property
    def soh(self) -> float:
        return self.get(self.SOH)

    @soh.setter
    def soh(self, value: float) -> None:
        self.set(self.SOH, value)

    @property
    def capacity(self) -> float:
        return self.get(self.CAPACITY)

    @capacity.setter
    def capacity(self, value: float) -> None:
        self.set(self.CAPACITY, value)

    @property
    def temperature(self) -> float:
        return self.get(self.TEMPERATURE)

    @temperature.setter
    def temperature(self, value: float) -> None:
        self.set(self.TEMPERATURE, value)

    @property
    def solar_thermal_load(self) -> float:
        return self.get(self.SOLAR_THERMAL_LOAD)

    @solar_thermal_load.setter
    def solar_thermal_load(self, value: float) -> None:
        self.set(self.SOLAR_THERMAL_LOAD, value)

    @property
    def hvac_thermal_power(self) -> float:
        return self.get(self.HVAC_THERMAL_POWER)

    @hvac_thermal_power.setter
    def hvac_thermal_power(self, value: float) -> None:
        self.set(self.HVAC_THERMAL_POWER, value)

    @property
    def ambient_temperature(self) -> float:
        return self.get(self.AMBIENT_TEMPERATURE)

    @ambient_temperature.setter
    def ambient_temperature(self, value: float) -> None:
        self.set(self.AMBIENT_TEMPERATURE, value)

    @property
    def max_charge_power(self) -> float:
        return self.get(self.MAX_CHARGE_POWER)

    @max_charge_power.setter
    def max_charge_power(self, value: float) -> None:
        self.set(self.MAX_CHARGE_POWER, value)

    @property
    def max_discharge_power(self) -> float:
        return self.get(self.MAX_DISCHARGE_POWER)

    @max_discharge_power.setter
    def max_discharge_power(self, value: float) -> None:
        self.set(self.MAX_DISCHARGE_POWER, value)

    @property
    def id(self) -> str:
        return 'SYSTEM' + str(self.get(self.SYSTEM_AC_ID)) + str(self.get(self.SYSTEM_DC_ID))

    @classmethod
    def sum_parallel(cls, system_states: []):
        system_state = SystemState(0, 0)
        for state in system_states:
            system_state.add(state)
        # average values
        size = len(system_states)
        system_state.divide_by(size, SystemState.AC_FULFILLMENT)
        system_state.divide_by(size, SystemState.TIME)
        system_state.divide_by(size, SystemState.DC_VOLTAGE_CIRCUIT)
        # system_state.divide_by(size, SystemState.DC_VOLTAGE)
        system_state.divide_by(size, SystemState.TEMPERATURE)
        # calculate capacity weighted soc and soh
        system_state.soc = 0
        system_state.soh = 0
        for state in system_states:
            system_state.soc += state.soc * state.capacity / system_state.capacity
            system_state.soh += state.soh * state.capacity / system_state.capacity
        if abs(system_state.ac_power) < max(system_state.max_charge_power, system_state.max_discharge_power) * 0.01:
            system_state.ac_fulfillment = 1
        else:
            system_state.ac_fulfillment = system_state.ac_power_delivered / system_state.ac_power
        return system_state

    @property
    def is_charge(self) -> bool:
        return self.ac_power > 0

    @classmethod
    def sum_serial(cls, states: []):
        raise Exception('Not implemented yet')



from configparser import ConfigParser
from datetime import datetime

from pytz import timezone

from simses.commons.config.simulation.simulation_config import SimulationConfig


class GeneralSimulationConfig(SimulationConfig):
    """
    General simulation configs
    """

    SECTION: str = 'GENERAL'

    START: str = 'START'
    END: str = 'END'
    TIME_FORMAT: str = '%Y-%m-%d %H:%M:%S'
    TIME_STEP: str = 'TIME_STEP'
    EXPORT_DATA: str = 'EXPORT_DATA'
    LOOP: str = 'LOOP'
    EXPORT_INTERVAL: str = 'EXPORT_INTERVAL'

    __UTC: timezone = timezone('UTC')

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def timestep(self) -> float:
        """Returns simulation timestep in s"""
        return float(self.get_property(self.SECTION, self.TIME_STEP))

    @property
    def start(self) -> float:
        """Returns simulation start timestamp"""
        date: str = self.get_property(self.SECTION, self.START)
        return self.__extract_utc_timestamp_from(date)

    @property
    def end(self) -> float:
        """Returns simulation end timestamp"""
        date: str = self.get_property(self.SECTION, self.END)
        return self.__extract_utc_timestamp_from(date)

    def __extract_utc_timestamp_from(self, date: str) -> float:
        date: datetime = datetime.strptime(date, self.TIME_FORMAT)
        return self.__UTC.localize(date).timestamp()

    @property
    def duration(self) -> float:
        """Returns simulation duration in s from __analysis_config file_name"""
        return self.end - self.start

    @property
    def loop(self) -> int:
        """Returns number of simulation loops"""
        return int(self.get_property(self.SECTION, self.LOOP))

    @property
    def export_interval(self) -> float:
        """Returns interval to write value to file"""
        return float(self.get_property(self.SECTION, self.EXPORT_INTERVAL))

    @property
    def export_data(self) -> bool:
        """Returns selection for data export True/False"""
        return self.get_property(self.SECTION, self.EXPORT_DATA) in ['True']



from configparser import ConfigParser

from simses.simulation.batch_processing import BatchProcessing


class PeakShavingCaseStudyBatchProcessing(BatchProcessing):

    def __init__(self):
        super().__init__(do_simulation=True, do_analysis=True)

    def __generate_dc_systems(self) -> dict:
        config_set: dict = dict()
        config_set['simses_lib_rfb'] = 'system_1,fix,lib\nsystem_1,fix,rfb\n'
        config_set['simses_lib_only'] = 'system_1,no_loss,lib_only\n'
        config_set['simses_rfb_only'] = 'system_1,no_loss,rfb_only\n'
        return config_set

    def _setup_config(self) -> dict:
        # generate config
        dc_systems: dict = self.__generate_dc_systems()
        # setup config
        configs: dict = dict()
        for name in dc_systems.keys():
            config: ConfigParser = ConfigParser()
            config.add_section('STORAGE_SYSTEM')
            config.set('STORAGE_SYSTEM', 'STORAGE_SYSTEM_DC', dc_systems[name])
            configs[name] = config
        # print configs
        print(dc_systems)
        print('Configured ' + str(len(configs)) + ' simulations')
        return configs

    def clean_up(self) -> None:
        pass


if __name__ == "__main__":
    batch_processing: BatchProcessing = PeakShavingCaseStudyBatchProcessing()
    batch_processing.run()
    print(type(batch_processing).__name__ + ' finished')



from simses.commons.state.system import SystemState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.generation.dc_generation import DcGeneration
from simses.system.dc_coupling.load.dc_load import DcLoad


class DcCoupling:

    """
    DcCouplings provide a possiblity to add a DC load or DC generation to the intermediate circuit of an AC storage
    system.
    """

    def __init__(self, dc_load: DcLoad, dc_generation: DcGeneration):
        self.__dc_load: DcLoad = dc_load
        self.__dc_generation: DcGeneration = dc_generation

    def update(self, time: float, state: SystemState) -> None:
        """
        In-place update of system state for dc power additional

        Parameters
        ----------
        time :
            current timestamp in s
        state :
            current system state
        """
        self.__dc_load.calculate_power(time)
        self.__dc_generation.calculate_power(time)
        state.dc_power_additional += self.get_power()

    def get_power(self) -> float:
        """

        Returns
        -------
        float:
            net power of DC load and DC generation in W
        """
        return self.__dc_generation.get_power() - self.__dc_load.get_power()

    def get_auxiliaries(self) -> [Auxiliary]:
        auxiliaries: [Auxiliary] = list()
        auxiliaries.extend(self.__dc_load.get_auxiliaries())
        auxiliaries.extend(self.__dc_generation.get_auxiliaries())
        return auxiliaries

    def close(self):
        self.__dc_load.close()
        self.__dc_generation.close()



from configparser import ConfigParser

from simses.commons.config.simulation.simulation_config import SimulationConfig


class HydrogenConfig(SimulationConfig):
    """
    Hydrogen storage specific configs
    """

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)
        self.__section: str = 'HYDROGEN'

    @property
    def soc(self) -> float:
        """Returns start soc from data_config file_name"""
        return float(self.get_property(self.__section, 'START_SOC'))

    @property
    def min_soc(self) -> float:
        """Returns min soc from data_config file_name"""
        return float(self.get_property(self.__section, 'MIN_SOC'))

    @property
    def max_soc(self) -> float:
        """Returns max soc from data_config file_name"""
        return float(self.get_property(self.__section, 'MAX_SOC'))

    # @property
    # def eol_electrolyzer(self) -> float:
    #     """Returns end of life criterion from data_config file_name"""
    #     return float(self.get_property(self.__section, 'EOL_ELECTROLYZER'))
    #
    # @property
    # def start_pressure_cathode_el(self) -> float:
    #     """Retruns start pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'START_PRESSURE_CATHODE_ELECTROLYZER'))
    #
    # @property
    # def desire_pressure_cathode_el(self) -> float:
    #     """Retruns desired pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'DESIRE_PRESSURE_CATHODE_ELECTROLYZER'))
    #
    # @property
    # def start_pressure_anode_el(self) -> float:
    #     """Retruns start pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'START_PRESSURE_ANODE_ELECTROLYZER'))
    #
    # @property
    # def desire_pressure_anode_el(self) -> float:
    #     """Retruns desired pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'DESIRE_PRESSURE_ANODE_ELECTROLYZER'))
    #
    # @property
    # def start_temperature_el(self) -> float:
    #     """Retruns start pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'START_TEMPERATURE_ELECTROLYZER'))
    #
    # @property
    # def desire_temperature_el(self) -> float:
    #     """Retruns desired pressure of cathode of electrolyzer"""
    #     return float(self.get_property(self.__section, 'DESIRE_TEMPERATURE_ELECTROLYZER'))
    #
    # @property
    # def serial_scale(self) -> float:
    #     """Returns a linear scaling factor of cell in order to simulate a serial lithium_ion connection"""
    #     return float(self.get_property(self.__section, 'CELL_SERIAL_SCALE'))
    #
    # @property
    # def parallel_scale(self) -> float:
    #     """Returns a linear scaling factor of cell in order to simulate a parallel lithium_ion connection"""
    #     return float(self.get_property(self.__section, 'CELL_PARALLEL_SCALE'))
    #
    # @property
    # def cell_types(self) -> [str]:
    #     """Returns cell chemistry for simulation"""
    #     return self.get_property(self.__section, 'CELL_TYPE').replace(' ', '').split(',')
    #
    # @property
    # def serial_battery_connection(self) -> int:
    #     """Returns number of serial connected batteries"""
    #     return int(self.get_property(self.__section, 'SERIAL_CIRCUIT'))
    #
    # @property
    # def parallel_battery_connection(self) -> int:
    #     """Returns number of parallel connected batteries"""
    #     return int(self.get_property(self.__section, 'PARALLEL_CIRCUIT'))
    #
    # @property
    # def simulate_each_cell(self) -> bool:
    #     """Returns boolean value for multithreading of batteries"""
    #     return self.get_property(self.__section, 'SIMULATE_EACH_CELL') in ['True']
    #
    # @property
    # def thermal_model_el(self) -> str:
    #     """Returns name of thermal model """
    #     return self.get_property(self.__section, 'THERMAL_MODEL')



from configparser import ConfigParser

from simses.commons.config.generation.analysis import AnalysisConfigGenerator
from simses.commons.config.generation.simulation import SimulationConfigGenerator
from simses.commons.config.simulation.battery import BatteryConfig
from simses.simulation.batch_processing import BatchProcessing
from simses.simulation.simbas.room_tool_reader import RoomToolReader
import os
import copy


class SimBAS(BatchProcessing):

    """
    This is the SimBAS example for BatchProcessing.
    """

    __CELL_EXT: str = '.xml'

    def __init__(self, use_room_tool: bool = True):
        super().__init__(do_simulation=True, do_analysis=True)
        self.__use_room_tool: bool = use_room_tool

    def _setup_config(self) -> dict:

        # Use case:
        #use_case = 'ecar'
        # use_case = 'eboat'
        use_case = 'hpc+bess'

        if use_case == 'ecar':
            config_file_name = 'simulation_SimBAS_Ecar'
            room_tool_file = 'report_auto.csv'
            ac_power: float = 105e3
        elif use_case == 'eboat':
            config_file_name = 'simulation_SimBAS_Eboot'
            room_tool_file = 'report_boot.csv'
            ac_power: float = 320000.0
        elif use_case == 'hpc+bess':
            config_file_name = 'simulation_SimBAS_HPC+BSS'
            room_tool_file = 'report_ladesäule.csv'
            ac_power: float = 320000.0


        profile_path = os.getcwd() + '\Profiles'

        config_generator: SimulationConfigGenerator = SimulationConfigGenerator()
        config_generator.load_default_config()
        config_generator.load_specific_config(config_file_name)

        # Set profile direction
        config_generator.set_profile_direction(profile_path)

        # generating config options
        config_generator.clear_storage_technology()
        dcdc_1: str = config_generator.add_fix_efficiency_dcdc(1.0)
        acdc_1: str = config_generator.add_no_loss_acdc()
        housing_1: str = config_generator.add_no_housing()
        hvac_1: str = config_generator.add_no_hvac()
        # generating storage systems
        config_generator.clear_storage_system_ac()
        config_generator.clear_storage_system_dc()
        # setting up multiple configurations with manual naming of simulations

        config_set: dict = dict()
        count: int = 0

        room_tool_reader: RoomToolReader = RoomToolReader(room_tool_file)
        room_tool_entries = room_tool_reader.get_data_report()
        for current_number in range(len(room_tool_entries)):

            cell = room_tool_entries["Model"][current_number]
            serial, parallel = 1, 1
            current_config_generator = copy.deepcopy(config_generator)

            cell_type: str = 'IseaCellType;' + cell + '_00001'# + self.__CELL_EXT
            # serial, parallel = room_tool_reader.get_battery_scale()
            serial = int(room_tool_entries["Cells in series"][current_number])
            parallel = int(room_tool_entries["Cells in parallel"][current_number])
            energy = int(room_tool_entries["Energy (Wh)"][current_number])
            voltage_ic = int(room_tool_entries["Nom. module voltage (V)"][current_number])

            # room_tool_reader: RoomToolReader = RoomToolReader(room_tool_file, cell=cell)
            # print(serial, parallel)
            # capacity = room_tool_reader.get_energy()
            # voltage_ic = room_tool_reader.get_nominal_voltage()
            storage = current_config_generator.add_lithium_ion_battery(capacity=energy, cell_type=cell_type)
            ac_system_1: str = current_config_generator.add_storage_system_ac(ac_power, voltage_ic, acdc_1, housing_1,
                                                                          hvac_1)
            current_config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage)
            count += 1
            config: ConfigParser = current_config_generator.get_config()
            # Attention: SimSES can only handle ONE serial/parallel config for ALL batteries
            # config.add_section('BATTERY')
            config.set(BatteryConfig.SECTION, BatteryConfig.CELL_SERIAL_SCALE, str(serial))
            config.set(BatteryConfig.SECTION, BatteryConfig.CELL_PARALLEL_SCALE, str(parallel))
            config_set['storage_' + str(count)] = config
            # for section in config.sections():
            #     print(section)
            #     print(dict(config.items(section)))
            # config_generator.show()
        return config_set

    def _analysis_config(self) -> ConfigParser:
        config_generator: AnalysisConfigGenerator = AnalysisConfigGenerator()
        config_generator.print_results(False)
        config_generator.do_plotting(True)
        config_generator.do_batch_analysis(True)
        return config_generator.get_config()

    def clean_up(self) -> None:
        pass

    def __read_cell_config(self, filename: str, delimiter: str = ',') -> [[str]]:
        cell_config: [[str]] = list()
        with open(filename, 'r', newline='') as file:
            for line in file:
                line: str = line.rstrip()
                if not line or line.startswith('#') or line.startswith('"'):
                    continue
                cell_config.append(line.split(delimiter))
        return cell_config


if __name__ == "__main__":
    batch_processing: BatchProcessing = SimBAS()
    batch_processing.run()
    batch_processing.clean_up()



import os
from configparser import ConfigParser
from datetime import datetime, timezone

import pytest

from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.profile.power.file import FilePowerProfile
from simses.commons.state.system import SystemState
from simses.commons.utils.utilities import remove_file
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.logic.energy_management.strategy.basic.peak_shaving_simple import \
    SimplePeakShaving

DELIMITER: str = ','
HEADER: str = '# Unit: W'
system_state: SystemState = SystemState(0,0)
START: int = 0
END: int = 5
STEP: int = 1
MAX_POWER: int = 3
FILE_NAME: str = os.path.join(os.path.dirname(__file__), 'mockup_power_profile.csv')


def create_file():
    with open(FILE_NAME, mode='w') as file:
        file.write(HEADER + '\n')
        value = START
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
            value += 1

    power_profile = FilePowerProfile(config=create_general_config(), filename=FILE_NAME,
                                          delimiter=DELIMITER, scaling_factor=1)
    return power_profile

def create_general_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('GENERAL')
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set('GENERAL', 'START', date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set('GENERAL', 'END', date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set('GENERAL', 'TIME_STEP', str(STEP))
    return GeneralSimulationConfig(config=conf)

def create_ems_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('ENERGY_MANAGEMENT')
    conf.set('ENERGY_MANAGEMENT', 'MAX_POWER', str(MAX_POWER))
    return EnergyManagementConfig(config=conf)


@pytest.fixture(scope='module')
def uut():
    power_profile = create_file()
    uut = SimplePeakShaving(power_profile=power_profile, ems_config=create_ems_config())
    yield uut
    uut.close()
    remove_file(FILE_NAME)

# time = power
@pytest.mark.parametrize('time, soc, power_offset, result',
                         [
                            (0, 0, 0, 3),
                            (0, 1, 0, 0),
                            (1, 0, 0, 2),
                            (1, 0, 1, 1),
                            (2, 0, 0, 1),
                            (2, 1, 0, 0),
                            (3, 0, 0, 0),
                            (3, 1, 0, 0),
                            (3, 0, 1, -1),
                            (3, 1, -1, 0),
                            (4, 0, 0, -1),
                            (4, 1, 0, -1),
                            (5, 0, 0, -2),
                            (5, 0, 2, -4)
                         ]
                         )
def test_next(time, soc, power_offset, result, uut):
    power_offset = power_offset
    system_state.soc = soc
    assert abs(uut.next(time, system_state, power_offset) - result) <= 1e-10


from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class EnergyManagementDataConfig(DataConfig):
    """class top read Energy Management data path"""

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'ENERGY_MANAGEMENT_DATA'

    @property
    def energy_management_data(self) -> str:
        """Returns directory of energy management data files"""
        return self.get_data_path(self.get_property(self.__section, 'ENERGY_MANAGEMENT_DATA_DIR'))

    @property
    def linear_efficiency_notton_acdc_file(self) -> str:
        """Returns filename for coefficients of linearized efficiency of Notton ACDC converter"""
        return self.energy_management_data + self.get_property(self.__section, 'LINEARIZED_EFFICIENCY_Notton_ACDC_FILE')

    @property
    def linear_efficiency_sony_lfp_file(self) -> str:
        """Returns filename for coefficients of linearized efficiency of Sony LFP"""
        return self.energy_management_data + self.get_property(self.__section, 'LINEARIZED_EFFICIENCY_Sony_LFP_FILE')

    @property
    def linear_cyc_degradation_sony_lfp_file(self) -> str:
        """Returns filename for coefficients of linearized cyclic degradation of Sony LFP"""
        return self.energy_management_data + self.get_property(self.__section,
                                                               'LINEARIZED_CYC_DEGRADATION_Sony_LFP_FILE')

    @property
    def linear_cal_degradation_sony_lfp_file(self) -> str:
        """Returns filename for coefficients of linearized calendar degradation of Sony LFP"""
        return self.energy_management_data + self.get_property(self.__section,
                                                               'LINEARIZED_CAL_DEGRADATION_Sony_LFP_FILE')



from simses.commons.constants import Hydrogen
from simses.system.auxiliary.auxiliary import Auxiliary


class WaterHeating(Auxiliary):

    def __init__(self):
        super().__init__()
        self.__heating_power = 0  # W
        self.__heating_efficiency = 1
        self.__delta_temperature = 0

    def ac_operation_losses(self) -> float:
        if self.__delta_temperature >= 0:  # electricity consumption only in case of heating
            return self.__heating_power / self.__heating_efficiency
        else:
            return 0

    def calculate_heating_power(self, water_flow, delta_temperature):
            self.__heating_power = Hydrogen.HEAT_CAPACITY_WATER * water_flow * Hydrogen.MOLAR_MASS_WATER * \
                                   delta_temperature
            self.__delta_temperature = delta_temperature

    def get_heating_power(self):
        return self.__heating_power


import math as m

import numpy as np
import pandas as pd

IDEAL_GAS_CONST: float = 8.314462  # J/(mol K)
FARADAY_CONST: float = 96485.3321  # As/mol

# constants
A_CATHODE = 2.4  # bar cm²/A  -> angaben, Abhänigkeit von Dicke
A_ANODE = 2.8  # bar cm²/A
R_ele = 0.096  # Ohm cm²
U_rev_STP = 1.23  # V
LAMBDA = 25  # degree of humidification
THIKNESS_NAFION = 200  # um

# parameters
ALPHA_array = np.array([0.02803, 0.02483, 0.02344, 0.02232, 0.02141, 0.02038])  # V
i_0_array = np.array([21.8, 5.71, 4.15, 3.36, 2.95, 1.97])* 10 **(-8)  # A/cm2
#R_mem_array = np.array([0.159, 0.140, 0.123, 0.109, 0.098, 0.088])  #  ohm cm²  direkte Parameter aus Q1

# input variables
T_array = [30.0, 40.0, 50.0, 60.0, 70.0, 80.0]  # °C
p_anode_array = np.linspace(0, 50, 13)  # barg  relative pressure compared to 1 bar ambient pressure
p_cathode_array = np.linspace(0, 50, 13)  # barg  relative pressure compared to 1 bar ambient pressure

U_rev_array = np.ones(len(T_array))
for i in range(len(T_array)):
    U_rev_array[i] = 1.5184 - 1.5421 * 10 ** (-3) * (T_array[i] + 273.15) + 9.523 * 10 ** (-5) * (T_array[i] + 273.15) * m.log(T_array[i] + 273.15) \
                     + 9.84 * 10 ** (-8) * (T_array[i] + 273.15) ** 2

# creating current density array with increasing step size to have more points in non linear area of the polariszation curve
current_density_array = []
current_density_array.append(0)
x = 0
i = 1
while x<3:
    x = x + 0.002*i**1.5
    current_density_array.append(x)
    i = i+1
current_density_array = np.array(current_density_array)
current_density_array[-1] = 3

pressure_sat_h2o_array = np.ones(len(T_array))  # bar
for i in range(len(T_array)):
    pressure_sat_h2o_array[i] = 10 ** (-2.1794 + 0.02953 * T_array[i] - 9.1837 * 10 ** -5 * T_array[i] ** 2 + 1.4454 * 10 ** -7 *
                            T_array[i] ** 3)

conductivity_nafion = np.ones(len(T_array))
for i in range(len(T_array)):
    conductivity_nafion[i] = (0.005139 * LAMBDA - 0.00326) * m.exp(1268 * (1 / 303 - 1 / (T_array[i]+273.15)))  # S/cm 0.14

R_mem_array = np.ones(len(T_array))
for i in range(len(T_array)):
    R_mem_array[i] = THIKNESS_NAFION * 10 ** (-4) / conductivity_nafion[i]  # Ohm cm2



temperature = []
p_anode = []
p_cathode = []
current_density = []
cell_voltage = []
alpha = []
i_0 = []

i=0

for i in range(0,len(T_array)):
    for j in range(0, len(p_anode_array)):
        for k in range(0, len(p_cathode_array)):
            for l in range(0, len(current_density_array)):
                # fill look-up-table with input values
                temperature.append(T_array[i])
                alpha.append(ALPHA_array[i])
                i_0.append(i_0_array[i])
                p_anode.append(p_anode_array[j])
                p_cathode.append(p_cathode_array[k])
                current_density.append(current_density_array[l])

                # calculate partial pressures of hydrogen and oxygen for calculation of Nernst-voltage without currentdependensy
                p_h2 = (1+p_cathode_array[k]) - pressure_sat_h2o_array[i]
                p_o2 = (1+p_anode_array[j]) - pressure_sat_h2o_array[i]
                p_h2_ref = 1  # bar
                p_o2_ref = 1  # bar

                # Nernst-voltage
                Nernst_Voltage = U_rev_array[i] + IDEAL_GAS_CONST * (T_array[i] + 273.15) / (2 * FARADAY_CONST) * \
                                 m.log((p_o2 / p_o2_ref)**(1/2) * p_h2 / p_h2_ref)
                #Nernst_Voltage = 1.23 + IDEAL_GAS_CONST * (T_array[i] + 273.15) / (2 * FARADAY_CONST) * \
                                 #m.log((p_o2 / p_o2_ref) ** (1 / 2) * p_h2 / p_h2_ref)

                # Activation-voltage
                if current_density_array[l] == 0:
                    Act_Voltage = 0
                else:
                    Act_Voltage = ALPHA_array[i] * m.log(current_density_array[l] / i_0_array[i])

                # ohmic voltage
                Ohm_Voltage = (R_mem_array[i] + R_ele) * current_density_array[l]

                # cell voltage
                cell_voltage.append(Nernst_Voltage + Act_Voltage + Ohm_Voltage)
                #cell_voltage.append(Nernst_Voltage)

temperature = np.array(temperature)
p_anode = np.array(p_anode)
p_cathode = np.array(p_cathode)
current_density = np.array(current_density)
cell_voltage = np.array(cell_voltage)
alpha = np.array(alpha)
i_0 = np.array(i_0)

# cell voltage lookup field - polariszation curve
cell_voltage_lookup_input = np.array([temperature, p_anode, p_cathode, current_density])
cell_voltage_lookup_output = np.array(cell_voltage)

dict1 = {'temperature in °C' : temperature, 'P anode in  bar' : p_anode, 'P cathode in bar' : p_cathode,
         'currentdensity in A/cm2' : current_density, 'cellvoltage in V' : cell_voltage}
df1 = pd.DataFrame(dict1)
df1.to_csv('cell_voltage_lookup_no_current_influence_on_part_pressure.csv', index=False)

# cell currentdensity lookup field - power curve
cell_power_density = cell_voltage * current_density

cell_current_density_lookup_input = np.array([temperature, p_anode, p_cathode, cell_power_density])
cell_current_density_lookup_output = current_density

dict2 = {'temperature in °C' : temperature, 'P anode in  bar' : p_anode, 'P cathode in bar' : p_cathode,
         'powerdensity in W/cm2' : cell_power_density, 'currentdensity in A/cm2' : current_density, 'alpha in V': alpha, 'i_0 in A/cm2': i_0}
df2 = pd.DataFrame(dict2)
df2.to_csv('cell_currentdensity_lookup_no_current_influence_on_part_pressure_with_alpha.csv', index=False)

# polarizationcurve_interpol = scipy.interpolate.LinearNDInterpolator(cell_voltage_lookup_input.T, cell_voltage_lookup_output)
# voltage1 = polarizationcurve_interpol(30.8,10.2,40.6,2.22)
# print(voltage1)



from math import sqrt

from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.pump_algorithm.abstract_pump_algorithm import PumpAlgorithm
from simses.technology.redox_flow.electrochemical.control.redox_control_system import RedoxControlSystem
from simses.technology.redox_flow.electrochemical.abstract_electrochemical import ElectrochemicalModel
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class RintModel(ElectrochemicalModel):
    """Equivalent circuit model that calculates the current and voltage of the redox flow stack module based on an
    internal resistance."""

    def __init__(self, stack_module: StackModule, control_system: RedoxControlSystem,
                 electrolyte_system: ElectrolyteSystem, pump_algorithm: PumpAlgorithm):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module: StackModule = stack_module
        self.__control_system: RedoxControlSystem = control_system
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system
        self.__pump_algorithm = pump_algorithm
        self.__FARADAY = self.__electrolyte_system.FARADAY
        self.__concentration_V = self.__electrolyte_system.get_vanadium_concentration()  # mol/m^3

    def update(self, time: float, redox_flow_state: RedoxFlowState) -> None:
        ocv: float = self.__stack_module.get_open_circuit_voltage(redox_flow_state)  # V
        redox_flow_state.open_circuit_voltage = ocv
        redox_flow_state.power = redox_flow_state.power_in
        rint: float = self.__stack_module.get_internal_resistance(redox_flow_state)  # Ohm

        # power check
        if abs(redox_flow_state.power) > self.__stack_module.get_max_power():
            if redox_flow_state.power < 0:
                redox_flow_state.power = -self.__stack_module.get_max_power()
            else:
                redox_flow_state.power = self.__stack_module.get_max_power()
        if (ocv ** 2 + 4 * rint * redox_flow_state.power) < 0:
            redox_flow_state.voltage = ocv
        else:
            redox_flow_state.voltage = (ocv + sqrt(ocv ** 2 + 4 * rint * redox_flow_state.power)) / 2
        self.__log.debug('System OCV: ' + str(ocv) + ', system voltage: ' + str(redox_flow_state.voltage))

        # Check temperature
        if not self.__control_system.check_temperature(redox_flow_state):
            self.__log.error('Electrolyte temperature is not in range. Destruction of the system is possible. HVAC must '
                            'be adjusted.')

        # Voltage check
        if not self.__control_system.check_voltage_in_range(redox_flow_state):
            self.__log.warn('Voltage ' + str(redox_flow_state.voltage) + ' V is not in range and is adjusted to ' +
                            str(self.__control_system.voltage_in_range(redox_flow_state)))
            redox_flow_state.voltage = self.__control_system.voltage_in_range(redox_flow_state)

        redox_flow_state.current = (redox_flow_state.voltage - ocv) / rint

        # Current check
        if not self.__control_system.check_current_in_range(redox_flow_state, time):
            self.__log.warn('Current ' + str(redox_flow_state.current) + ' A is not in range. Max value ' + str(
                self.__control_system.get_max_current(redox_flow_state, time)) + 'A, min value ' + str(
                self.__control_system.get_min_current(redox_flow_state, time)) + ' A.')
            if redox_flow_state.current > 0:
                redox_flow_state.current = self.__control_system.get_max_current(redox_flow_state, time)
            else:
                redox_flow_state.current = self.__control_system.get_min_current(redox_flow_state, time)
            redox_flow_state.voltage = redox_flow_state.current * rint + ocv
            self.__log.debug('Current after control system check: ' + str(redox_flow_state.current) + ', Voltage: ' +
                             str(redox_flow_state.voltage) + ' V.')

        delta_soc, delta_soc_stack, soc_stack = self.__calculate_soc(time, redox_flow_state, self.__stack_module)

        # SOC check
        if not self.__control_system.check_soc_in_range(redox_flow_state, delta_soc, delta_soc_stack):
            self.__log.warn('RFB should not be charged/discharged due to SOC restrictions. SOC: ' + str(
                redox_flow_state.soc) + ', power (' + str(redox_flow_state.power) + ') is set to 0')
            redox_flow_state.current = 0.0
            redox_flow_state.voltage = ocv
            redox_flow_state.flow_rate_catholyte = 0.0
            redox_flow_state.flow_rate_anolyte = 0.0
            delta_soc, delta_soc_stack, soc_stack = self.__calculate_soc(time, redox_flow_state, self.__stack_module)

        redox_flow_state.power = redox_flow_state.current * redox_flow_state.voltage
        redox_flow_state.power_loss = rint * redox_flow_state.current ** 2
        redox_flow_state.internal_resistance = rint
        redox_flow_state.soc_stack = soc_stack + delta_soc_stack
        redox_flow_state.soc = redox_flow_state.soc + delta_soc
        self.__log.debug('New SOC: ' + str(redox_flow_state.soc))

        # Battery fulfillment
        self.__control_system.battery_fulfillment_calc(redox_flow_state)

        # setting max power
        self.__control_system.set_max_power(redox_flow_state)

    def __calculate_soc(self, time: float, redox_flow_state: RedoxFlowState, stack_module: StackModule):
        """
        Calculates the state-of-charge (SOC) of the system and in the cells of the stack module.

        Parameters
        ----------
         time : float
            Current simulation time in s.
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.
        stack_module : StackModule
            Type of redox flow battery stack module.

        Returns
        -------
            delta_soc in p.u.
            delta_soc_stack in p.u.
        """
        cell_num_stack = stack_module.get_cell_per_stack()
        self_discharge_current: float = stack_module.get_self_discharge_current(redox_flow_state)
        soc_stack = redox_flow_state.soc_stack
        delta_soc_stack = 0.0
        soc = redox_flow_state.soc
        capacity_amps = redox_flow_state.capacity * 3600 / self.__stack_module.get_nominal_voltage_cell()  # As

        if redox_flow_state.flow_rate_anolyte == 0 or redox_flow_state.flow_rate_catholyte == 0:
            if redox_flow_state.current == 0:
                self_discharge_current = 0   # for no self-discharge during standby
            delta_soc_stack = (redox_flow_state.current * cell_num_stack * stack_module.get_serial_scale() -
                               self_discharge_current) * (time - redox_flow_state.time) / (self.__FARADAY *
                               self.__concentration_V * stack_module.get_stacks_electrolyte_volume())
            if (soc_stack + delta_soc_stack) < 0.001:
                # soc_stack can't be 0, because then some formula become invalid (e.g. OCV calculation)
                self.__log.warn('Stack is totally discharged.')
                self_discharge_current = 0
                delta_soc_stack = 0.0
        delta_soc = ((redox_flow_state.current * cell_num_stack * stack_module.get_serial_scale() -
                      self_discharge_current) * (time - redox_flow_state.time) / capacity_amps)
        self.__log.debug('Current: ' + str(redox_flow_state.current) + ' , SD-Current: ' +
                         str(self_discharge_current/(cell_num_stack * stack_module.get_serial_scale())))
        if not (self.__pump_algorithm.get_flow_rate_anolyte(redox_flow_state) == 0 or
                self.__pump_algorithm.get_flow_rate_catholyte(redox_flow_state) == 0):
            if (soc + delta_soc) < 0.001:
                delta_soc_stack = 0
            else:
                delta_soc_stack = delta_soc
            soc_stack = redox_flow_state.soc
        return delta_soc, delta_soc_stack, soc_stack

    def close(self) -> None:
        super().close()
        self.__log.close()
        self.__stack_module.close()
        self.__control_system.close()
        self.__electrolyte_system.close()



from simses.commons.constants import Hydrogen
from simses.technology.hydrogen.hydrogen_storage.pipeline.pipeline import Pipeline


class SimplePipeline(Pipeline):

    """calculates total mass of produced and injected hydrogen in kg"""

    def __init__(self, storage_pressure: float):
        super().__init__()
        self.__injected_hydrogen = 0  # kg
        self.__storage_pressure = storage_pressure

    def calculate_inijected_hydrogen(self, time_diff: float, hydrogen_outflow: float) -> None:
        self.__injected_hydrogen = time_diff * hydrogen_outflow * 2.0 * Hydrogen.MOLAR_MASS_HYDROGEN # kg

    def calculate_soc(self, time_diff: float, hydrogen_net_flow: float) -> None:
        pass

    def get_injected_hydrogen(self) -> float:
        return self.__injected_hydrogen

    def get_tank_pressure(self) -> float:
        return self.__storage_pressure

    def get_soc(self) -> float:
        return 0.5

    def get_capacity(self) -> float:
        return 1000000

    def close(self) -> None:
        pass



from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.plotting.energy_plotting import EnergyPlotting


class SunburstDiagram(EnergyPlotting):
    """
    SunburstDiagram handles the backend data processing for the sunburst_diagram method of PlotlyPlotting.
    This class dynamically creates labels, parents, and values of energetic quantities as per the
    BESS configuration in SimSES.
    """

    title_extension: str = 'Loss distribution'
    LABELS: str = 'LABELS'
    PARENTS: str = 'PARENTS'

    def __init__(self, data: SystemData, title: str, path: str):
        super().__init__(data, path)
        self.__title = title + self.title_extension
        self.__figures: list = list()

    def create_parameters(self) -> dict:
        """
        Creates parameters (labels, parents, and values) required to plot the Sunburst Diagram which depicts energetic
        quantites such as losses and charged/discharged energy for the BESS
        :param categories: dict of dynamically created energy quantities
        :return: parameters (labels, parents, and values) for the Sunburst Diagram as dict
        """
        # Compute quantities
        data: SystemData = self.get_data()
        transformer_losses_ch = 0
        acdc_converter_losses_ch = data.ac_pe_charging_energy - data.dc_power_charging_energy
        dcdc_converter_losses_ch = data.dc_power_charging_energy - data.dc_power_storage_charging_energy
        storage_technology_losses_ch = data.storage_technology_loss_energy/2

        transformer_losses_dch = 0
        acdc_converter_losses_dch = data.total_pe_losses_energy - acdc_converter_losses_ch
        dcdc_converter_losses_dch = abs(data.dc_power_discharging_energy - data.dc_power_storage_discharging_energy)
        storage_technology_losses_dch = data.storage_technology_loss_energy/2

        if abs(data.total_aux_losses_energy) < 1e-5:
            aux_energy_ch = 0
            aux_energy_dch = 0
        else:
            aux_energy_ch = data.aux_energy_charging
            aux_energy_dch = data.total_aux_losses_energy - data.aux_energy_charging

        # Create plot attributes
        categories: dict = self.create_categories()
        labels = list()
        parents = list()
        values = list()

        labels_length = len(labels)

        # Grid (Charging Energy) + Initial Energy Content as first parent
        core_parent_label = categories[self.COMPONENTS_CH][self.GRID] + ' + ' + categories[self.COMPONENTS_CH][self.INITIAL_ENERGY_CONTENT]
        labels.append(core_parent_label)
        values.append(data.charge_energy + data.initial_energy_content)

        labels_length_change = len(labels) - labels_length
        parents.extend(labels_length_change * [""])

        # Grid (Charging Energy) as parent of Grid (Discharging Energy) and Losses (Charging) and Losses (Discharging)
        labels_length = len(labels)
        labels.append(categories[self.COMPONENTS_DCH][self.GRID])
        values.append(data.discharge_energy)
        if abs(data.total_aux_losses_energy) >= 1e-5:
            labels.append(categories[self.COMPONENTS_CH][self.AUXILIARY])
            values.append(aux_energy_ch)
            labels.append(categories[self.COMPONENTS_DCH][self.AUXILIARY])
            values.append(aux_energy_dch)
        labels.append(self.LOSSES_CH)
        values.append(transformer_losses_ch + acdc_converter_losses_ch +
                      dcdc_converter_losses_ch + storage_technology_losses_ch)
        labels.append(self.LOSSES_DCH)
        values.append(transformer_losses_dch + acdc_converter_losses_dch +
                      dcdc_converter_losses_dch + storage_technology_losses_dch)
        labels.append(categories[self.COMPONENTS_CH][self.FINAL_ENERGY_CONTENT])
        values.append(data.final_energy_content)

        labels_length_change = len(labels) - labels_length
        parents.extend(labels_length_change * [core_parent_label])

        # Losses (Charging) as parent of component-wise charging losses
        labels_length = len(labels)
        labels.append(categories[self.LOSSES_CH][self.TRANSFORMER])
        values.append(transformer_losses_ch)
        labels.append(categories[self.LOSSES_CH][self.ACDC_CONVERTER])
        values.append(acdc_converter_losses_ch)
        labels.append(categories[self.LOSSES_CH][self.DCDC_CONVERTER])
        values.append(dcdc_converter_losses_ch)
        labels.append(categories[self.LOSSES_CH][self.STORAGE_TECHNOLOGY])
        values.append(storage_technology_losses_ch)

        labels_length_change = len(labels) - labels_length
        parents.extend(labels_length_change * [self.LOSSES_CH])

        # Losses (Discharging) as parent of component-wise discharging losses
        labels_length = len(labels)
        labels.append(categories[self.LOSSES_DCH][self.TRANSFORMER])
        values.append(transformer_losses_dch)
        labels.append(categories[self.LOSSES_DCH][self.ACDC_CONVERTER])
        values.append(acdc_converter_losses_dch)
        labels.append(categories[self.LOSSES_DCH][self.DCDC_CONVERTER])
        values.append(dcdc_converter_losses_dch)
        labels.append(categories[self.LOSSES_DCH][self.STORAGE_TECHNOLOGY])
        values.append(storage_technology_losses_dch)

        labels_length_change = len(labels) - labels_length
        parents.extend(labels_length_change * [self.LOSSES_DCH])

        parameters = dict()
        parameters[self.LABELS] = labels
        parameters[self.PARENTS] = parents
        parameters[self.VALUES] = values

        return parameters








from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class BatteryDataConfig(DataConfig):

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'BATTERY_DATA'

    @property
    def cell_data_dir(self) -> str:
        """Returns directory of cell data files"""
        return self.get_data_path(self.get_property(self.__section, 'CELL_DATA_DIR'))

    @property
    def lfp_sony_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of sony LFP"""
        return self.cell_data_dir + self.get_property(self.__section, 'LFP_SONY_OCV_FILE')

    @property
    def lfp_sony_degradation_capacity_file(self) -> str:
        """Returns filename for open circuit voltage of sony LFP"""
        return self.cell_data_dir + self.get_property(self.__section, 'LFP_SONY_DEGRADATION_CAPACITY_FILE')

    @property
    def lfp_sony_degradation_resistance_file(self) -> str:
        """Returns filename for open circuit voltage of sony LFP"""
        return self.cell_data_dir + self.get_property(self.__section, 'LFP_SONY_DEGRADATION_RESISTANCE_FILE')

    @property
    def lfp_sony_rint_file(self) -> str:
        """Returns filename for internal resistance of sony LFP"""
        return self.cell_data_dir + self.get_property(self.__section, 'LFP_SONY_RINT_FILE')

    @property
    def nca_panasonicNCR_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of panasonic NCA"""
        return self.cell_data_dir + self.get_property(self.__section, 'NCA_PANASONICNCR_OCV_FILE')

    @property
    def nca_panasonicNCR_rint_file(self) -> str:
        """Returns filename for internal resistance of panasonic NCA"""
        return self.cell_data_dir + self.get_property(self.__section, 'NCA_PANASONICNCR_RINT_FILE')

    @property
    def nmc_molicel_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_OCV_FILE')

    @property
    def nmc_molicel_rint_file(self) -> str:
        """Returns filename for internal resistance of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_RINT_FILE')

    @property
    def nmc_samsung94test_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of NMC 94Ah SamsungLabTest"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Samsung94AhTest_OCV_FILE')

    @property
    def nmc_samsung94test_rint_file(self) -> str:
        """Returns filename for internal resistance of NMC 94Ah SamsungLabTest"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Samsung94AhTest_RINT_FILE')

    @property
    def nmc_molicel_capacity_cal_file(self) -> str:
        """Returns filename for parameters for calendar aging of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_CAPACITY_CAL_FILE')

    @property
    def nmc_molicel_ri_cal_file(self) -> str:
        """Returns filename for parameters for calendar resistance increase of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_RI_CAL_FILE')

    @property
    def nmc_molicel_capacity_cyc_file(self) -> str:
        """Returns filename for parameters for cyclic aging of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_CAPACITY_CYC_FILE')

    @property
    def nmc_molicel_ri_cyc_file(self) -> str:
        """Returns filename for parameters for cyclic resistance increase of NMC Molicel"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Molicel_RI_CYC_FILE')

    @property
    def nmc_sanyo_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of sony Sanyo NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Sanyo_OCV_FILE')

    @property
    def nmc_sanyo_rint_file(self) -> str:
        """Returns filename for internal resistance of Sanyo NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Sanyo_RINT_FILE')

    @property
    def nmc_samsung_120ah_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of ads-tec Sanyo NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Samsung_120Ah_OCV_FILE')

    @property
    def nmc_samsung_120ah_rint_file(self) -> str:
        """Returns filename for internal resistance of ads-tec NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Samsung_120Ah_RINT_FILE')

    @property
    def nmc_samsung_120ah_capacity_cal_file(self) -> str:
        """Returns filename for calendar degradation of ads-tec NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Samsung_120Ah_CAPACITY_CAL_FILE')

    @property
    def nmc_akasol_akm_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of Akasol-AKM NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Akasol_AKM_OCV_FILE')

    @property
    def nmc_akasol_akm_rint_file(self) -> str:
        """Returns filename for internal resistance of Akasol-AKM NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Akasol_AKM_RINT_FILE')

    @property
    def nmc_akasol_oem_ocv_file(self) -> str:
        """Returns filename for open circuit voltage of Akasol-OEM NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Akasol_OEM_OCV_FILE')

    @property
    def nmc_akasol_oem_rint_file(self) -> str:
        """Returns filename for internal resistance of Akasol-OEM NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Akasol_OEM_RINT_FILE')

    @property
    def nmc_akasol_oem_current_file(self) -> str:
        """Returns filename for maximum current of Akasol-OEM NMC"""
        return self.cell_data_dir + self.get_property(self.__section, 'NMC_Akasol_OEM_CURRENT_FILE')

    @property
    def isea_cell_dir(self) -> str:
        """Returns dirname for isea cell files"""
        return self.get_data_path(self.get_property(self.__section, 'ISEA_CELL_DIR'))

    @property
    def lto_lmo_current_file(self) -> str:
        """Returns filename for lto_lmo cell files"""
        return self.cell_data_dir + self.get_property(self.__section, 'LTO_LMO_CURRENT_FILE')

    @property
    def lto_nmc_current_file(self) -> str:
        """Returns filename for lto_nmc cell files"""
        return self.cell_data_dir + self.get_property(self.__section, 'LTO_NMC_CURRENT_FILE')

    @property
    def sodium_ion_rint_green_rock_file(self) -> str:
        """Returns filename for internal resistance of sodium_ion Green Rock"""
        return self.cell_data_dir + self.get_property(self.__section, 'SODIUM_ION_RINT_GREEN_ROCK_FILE')

    @property
    def sodium_ion_ocv_green_rock_file(self) -> str:
        """Returns filename for open circuit voltage of sodium-ion Green Rock"""
        return self.cell_data_dir + self.get_property(self.__section, 'SODIUM_ION_OCV_GREEN_ROCK_FILE')


from simses.commons.log import Logger
from simses.commons.state.technology.hydrogen import HydrogenState
from simses.commons.utils.utilities import format_float
from simses.commons.config.simulation.hydrogen import HydrogenConfig


class HydrogenManagementSystem:

    def __init__(self, min_power_electroyzer: float, max_power_electrolyzer: float, max_power_fuel_cell: float, config: HydrogenConfig):
        self.__log: Logger = Logger(type(self).__name__)
        self.__min_charge_power: float = min_power_electroyzer
        self.__max_charge_power: float = max_power_electrolyzer  # max power is greater than nominal power
        self.__max_discharge_power: float = - max_power_fuel_cell  # max power is greater than nominal power
        self.__max_soc: float = config.max_soc
        self.__min_soc: float = config.min_soc

    def update(self, time: float, state: HydrogenState) -> None:
        """Updating power and fulfillment factor in order to check restrictions"""
        power_target = state.power
        state.power = self.__get_max_power(time, state)
        self.__set_fulfillment(power_target, state)

    def __get_max_power(self, time: float, state: HydrogenState) -> float:
        if not self.__power_in_range(state):
            self.__log.warn("Power out of range: " + format_float(state.power))
        power = max(min(state.power, self.__max_charge_power), self.__max_discharge_power)
        if 0 < power < self.__min_charge_power:
            power = 0
        soc_charge_power = self.__soc_max_charge_power(time, state)
        soc_discharge_power = self.__soc_max_discharge_power(time, state)
        if not soc_discharge_power < power < soc_charge_power:
            self.__log.warn('Power is limited due to SOC.')
        power = max(min(power, soc_charge_power), soc_discharge_power)
        return power

    def __power_in_range(self, state: HydrogenState):
        return self.__max_discharge_power < state.power < self.__max_charge_power # TODO Sign Error?

    def __soc_max_charge_power(self, time: float, state: HydrogenState) -> float:
        denergy = (self.__max_soc - state.soc) * state.capacity
        power = denergy / (time - state.time) * 3600
        return max(0.0, power)

    def __soc_max_discharge_power(self, time: float, state: HydrogenState) -> float:
        denergy = (state.soc - self.__min_soc) * state.capacity
        power = denergy / (time - state.time) * 3600
        return min(0.0, -power)

    def __set_fulfillment(self, power_target: float, state: HydrogenState) -> None:
        power_is = state.power
        if abs(power_is - power_target) < 1e-8:
            state.fulfillment = 1.0
        elif power_target == 0:
            self.__log.error('Power should be 0, but is ' + str(power_is) + ' W. Check BMS function')
            state.fulfillment = 0.0
        else:
            state.fulfillment = abs(power_is / power_target)
            if state.fulfillment < 0.0 or state.fulfillment > 1.0:
                self.__log.error('Fulfillment should be between 0 and 1 , but is ' +
                                 format_float(state.fulfillment) + ' p.u.. Check BMS function')

    def close(self) -> None:
        """Closing all resources"""
        self.__log.close()



from datetime import datetime
from random import Random

from pytz import timezone

from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.dc_coupling.load.dc_load import DcLoad


class DcBusChargingRandom(DcLoad):

    """
    DcBusChargingRandom acts as load profile for busses connected to the storage system at varying start and end
    times of each day.
    """

    __number_charging_stations: int = 0

    __START_CHARGE_HOUR_MEAN: float = 19.0
    __CHARGE_HOUR_VARIATION: float = 3.0

    __DURATION_CHARGE_MEAN: float = 3.0
    __CHARGE_DURATION_VARIATION: float = 1.0

    __UTC: timezone = timezone('UTC')

    def __init__(self, charging_power: float):
        super().__init__()
        self.__CHARGE_POWER: float = charging_power  # W
        self.__class__.__number_charging_stations += 1
        self.__random: Random = Random(3515723 * self.__class__.__number_charging_stations)
        self.__start_charge: float = 0.0
        self.__stop_charge: float = 0.0
        self.__day: int = 0
        self.__power: float = 0.0
        self.__charge_active: bool = False

    def __get_start_charging(self, date: datetime) -> float:
        tstmp: float = self.__random.uniform(self.__START_CHARGE_HOUR_MEAN - self.__CHARGE_HOUR_VARIATION,
                                     self.__START_CHARGE_HOUR_MEAN + self.__CHARGE_HOUR_VARIATION) * 3600.0
        minutes, seconds = divmod(tstmp, 60.0)
        hour, minutes = divmod(minutes, 60.0)
        date = date.replace(hour=int(hour), minute=int(minutes), second=int(seconds))
        return self.__UTC.localize(date).timestamp()

    def __get_charging_duration(self) -> float:
        return self.__random.uniform(self.__DURATION_CHARGE_MEAN - self.__CHARGE_DURATION_VARIATION,
                                     self.__DURATION_CHARGE_MEAN + self.__CHARGE_DURATION_VARIATION) * 3600.0

    def __set_charging_times(self, time: float) -> None:
        date = datetime.utcfromtimestamp(time)
        if self.__day != date.day and not self.__charge_active:
            self.__day = date.day
            self.__start_charge = self.__get_start_charging(date)
            self.__stop_charge = self.__start_charge + self.__get_charging_duration()

    def get_power(self) -> float:
        return self.__power

    def calculate_power(self, time: float) -> None:
        self.__set_charging_times(time)
        # hour = tstmp.hour + tstmp.minute / 60. + tstmp.second / 3600.
        if self.__start_charge < time < self.__stop_charge:
            self.__power = self.__CHARGE_POWER
            self.__charge_active = True
        else:
            self.__power = 0.0
            self.__charge_active = False

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass



import multiprocessing
import os
from configparser import ConfigParser
from multiprocessing import Queue

from simses.commons.utils.utilities import remove_all_files_from, create_directory_for
from simses.main import SimSES


class BatchSimulation(multiprocessing.Process):

    """
    BatchSimulation wraps the SimSES simulation into a multiprocessing Process in order to allow to run simulation
    in parallel. In addition, it is possible to run multiple threaded simulations within this process but this is
    strongly not recommended.
    """

    BATCH_DIR: str = 'batch/'

    def __init__(self, config_set: dict, printer_queue: Queue, path: str, batch_size: int = 1,
                 do_simulation: bool = True, do_analysis: bool = True, analysis_config: ConfigParser = ConfigParser()):
        super().__init__()
        self.__batch_dir: str = os.path.join(path, self.BATCH_DIR)
        create_directory_for(self.__batch_dir)
        remove_all_files_from(self.__batch_dir)
        self.__path: str = os.path.join(path)
        self.__batch_size: int = batch_size
        self.__printer_queue: Queue = printer_queue
        self.__config_set: dict = config_set
        self.__analysis_config: ConfigParser = analysis_config
        self.__do_simulation: bool = do_simulation
        self.__do_analysis: bool = do_analysis

    def __setup(self) -> [SimSES]:
        simulations: [SimSES] = list()
        for name, config in self.__config_set.items():
            remove_all_files_from(os.path.join(self.__path, name))
            simulations.append(SimSES(self.__path, name, do_simulation=self.__do_simulation,
                                      do_analysis=self.__do_analysis, simulation_config=config,
                                      analysis_config=self.__analysis_config, queue=self.__printer_queue,
                                      batch_dir=self.BATCH_DIR))
        return simulations

    def run(self):
        started: [SimSES] = list()
        simulations: [SimSES] = self.__setup()
        self.__check_simulation_names(simulations)
        for simulation in simulations:
            # print('\rStarting ' + simulation.name)
            simulation.start()
            started.append(simulation)
            if len(started) >= self.__batch_size:
                self.__wait_for(started)
                started.clear()
        self.__wait_for(started)

    @staticmethod
    def __wait_for(simulations: [SimSES]):
        for simulation in simulations:
            simulation.join()

    @staticmethod
    def __check_simulation_names(simulations: [SimSES]) -> None:
        names: [str] = list()
        for simulation in simulations:
            name: str = simulation.name
            if name in names:
                raise Exception(name + ' is not unique. Please check your simulation setup!')
            names.append(name)



import math
from datetime import datetime

import numpy
import numpy as np

from simses.analysis.data.abstract_data import Data
from simses.analysis.evaluation.abstract_evaluation import Evaluation
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.log import Logger
from simses.commons.utils.utilities import format_float


class TechnicalEvaluation(Evaluation):

    """
    TechnicalEvaluation is a special evaluation class for calculating technical KPIs, e.g. efficiency.
    """

    def __init__(self, data: Data, config: GeneralAnalysisConfig):
        super().__init__(data, config, config.technical_analysis)
        self.__log: Logger = Logger(type(self).__name__)

    def evaluate(self) -> None:
        self.append_result(EvaluationResult(Description.Technical.ROUND_TRIP_EFFICIENCY, Unit.PERCENTAGE, self.round_trip_efficiency))
        self.append_result(EvaluationResult(Description.Technical.MEAN_SOC, Unit.PERCENTAGE, self.mean_soc))
        self.append_result(EvaluationResult(Description.Technical.MAX_SOC, Unit.PERCENTAGE, self.max_soc))
        self.append_result(EvaluationResult(Description.Technical.MIN_SOC, Unit.PERCENTAGE, self.min_soc))
        self.append_result(EvaluationResult(Description.Technical.NUMBER_CHANGES_SIGNS, Unit.NONE, self.changes_of_sign))
        self.append_result(EvaluationResult(Description.Technical.RESTING_TIME_AVG, Unit.MINUTES, self.resting_times))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_CHANGES_SIGN, Unit.PERCENTAGE, self.energy_swapsign))
        self.append_result(EvaluationResult(Description.Technical.FULFILLMENT_AVG, Unit.PERCENTAGE, self.average_fulfillment))
        self.append_result(EvaluationResult(Description.Technical.REMAINING_CAPACITY, Unit.PERCENTAGE, self.capacity_remaining))
        self.append_result(EvaluationResult(Description.Technical.ENERGY_THROUGHPUT, Unit.KWH, self.energy_throughput))

    def plot(self) -> None:
        pass

    @property
    def round_trip_efficiency(self) -> float:

        """
        Calculates the round trip efficiency of the system/battery

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            round trip efficiency
        """
        data: Data = self.get_data()
        discharged = data.discharge_energy
        charged = data.charge_energy
        difference = data.energy_difference

        # eta_ohne_anpassung = 100 * (data.discharge_energy + data.energy_difference) / data.charge_energy
        # eta_formel_genau = 100 * 0.5*((b*math.sqrt(4*a*c+b**2))/c**2 + 2*a/c + (b/c)**2)

        if charged > 0.0 and discharged > 0.0:
            # return 100 * (data.discharge_energy + data.energy_difference) / data.charge_energy
            efficiency = 100 * 0.5 * ((difference * math.sqrt(
                4 * discharged * charged + difference ** 2)) / charged ** 2 + 2 * discharged / charged + (
                                                  difference / charged) ** 2)
        elif charged > 0.0:
            efficiency = 100 * difference / charged
            self.__log.warn('The storage was only charged. Thus, only a one-way efficiency could be calculated')
        elif discharged > 0.0:
            efficiency = 100 * -difference / discharged
            self.__log.warn('The storage was only discharged. Thus, only a one-way efficiency could be calculated')
        else:
            efficiency = numpy.nan
            self.__log.warn('Storage was neither charged nor discharged. Efficiency could not be calculated.')
        if not 0.0 < efficiency <= 100.0:
            self.__log.warn(Description.Technical.ROUND_TRIP_EFFICIENCY + ' should be between 0 % and 100 %, but is ' +
                             format_float(efficiency) + ' %. Perhaps your simulation time range is too short for '
                                                             'calculating an accurate round trip efficiency.')
        return efficiency

    @property
    def capacity_remaining(self) -> float:
        data: Data = self.get_data()
        return data.state_of_health[-1] * 100.0

    @property
    def energy_throughput(self) -> float:
        data: Data = self.get_data()
        return data.charge_energy + data.discharge_energy

    @property
    def max_soc(self) -> [float]:
        """
        Calculates the max SOC of the system/battery

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average soc
        """
        data: Data = self.get_data()
        return data.max_soc * 100.

    @property
    def mean_soc(self) -> [float]:
        """
        Calculates the mean SOC of the system/battery

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average soc
        """
        data: Data = self.get_data()
        return 100 * data.average_soc

    @property
    def min_soc(self) -> [float]:
        """
        Calculates the min SOC of the system/battery

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average soc
        """
        data: Data = self.get_data()
        return data.min_soc * 100.0

    @property
    def equivalent_full_cycles(self) -> float:
        """
        Calculates the number of full-equivalent cycles by dividing the amount of charged energy through the initial capacity

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            number of full-equivalent cycles
        """
        data: Data = self.get_data()
        return data.charge_energy / data.initial_capacity

    @property
    def depth_of_discharges(self) -> float:
        """
        Calculates the average depth of cycles in discharge direction

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average depth of cycles in discharge direction
        """
        data: Data = self.get_data()
        delta_soc = np.diff(data.soc)[abs(np.diff(data.soc)) > 1e-8]
        delta_soc_sign = np.sign(delta_soc)
        delta_soc_sign_diff = np.diff(delta_soc_sign)
        cycle_end = np.asarray(np.where(delta_soc_sign_diff != 0))
        cycle_start = cycle_end + 1
        cycle_start = np.insert(cycle_start, 0, 0)
        cycle_end = np.append(cycle_end, len(delta_soc) - 1)
        doc = np.asarray(
            [sum(delta_soc[cycle_start[counter]:cycle_end[counter] + 1]) for counter in range(0, len(cycle_start))])
        if len(doc) == 0:
            return 0
        if len(doc[doc < 0]) == 0:
            return 0
        doc_dis = abs(100 * doc[doc < 0].mean())
        return doc_dis

    @property
    def changes_of_sign(self) -> float:
        """
        Calculates the average number of changes of sign per day

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average number of changes of sign per day
        """
        data: Data = self.get_data()
        days_in_data = round(
            (datetime.fromtimestamp(data.time[-1]) - datetime.fromtimestamp(data.time[0])).total_seconds() / 86400)
        power_sign = np.sign(data.power)
        power_sign_nozero = power_sign[power_sign != 0]
        total_changes_of_sign = np.nansum(abs(np.diff(power_sign_nozero))) / 2
        if days_in_data > 0:
            daily_changes_of_sign = total_changes_of_sign / days_in_data
        else:
            daily_changes_of_sign = 'Less than one day simulated.'
        return daily_changes_of_sign

    @property
    def resting_times(self) -> float:
        """
        Calculates the average length of resting time of the system/battery

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average length of resting time in min
        """
        data: Data = self.get_data()
        resting_data = 1 * (abs(data.power) == 0)

        idx_begin = (np.diff(resting_data) > 0)
        idx_end = np.diff(resting_data) < 0
        if resting_data[0] == 1:
            idx_begin = np.insert(idx_begin, 0, True)
            idx_end = np.insert(idx_end, 0, False)
        if resting_data[-1] == True:
            idx_begin = np.append(idx_begin, False)
            idx_end = np.append(idx_end, True)
        times_length = np.where(idx_end)[0] - np.where(idx_begin)[0]
        timestep = data.time[1] - data.time[0]  # in seconds

        if len(times_length):
            resting_times_length = times_length.mean() * timestep / 60
        else:
            resting_times_length = 'Never in resting mode'

        return resting_times_length

    @property
    def energy_swapsign(self) -> float:
        """
        Calculates the average positive (charged) energy between changes of sign

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average charged energy between between changes of sign
        """
        data: Data = self.get_data()
        power_sign = np.sign(data.power)
        power = data.power
        timestep = data.time[1] - data.time[0]  # in seconds
        nom_capacity = data.initial_capacity

        pos_or_zero_value = np.where(power_sign >= 0)

        difference_greater_one = np.diff(pos_or_zero_value) !=1
        difference_greater_one = np.append(difference_greater_one[0], True) # append last pos. value
        last_pos_indexes = pos_or_zero_value[0][difference_greater_one]
        first_pos_indexes = pos_or_zero_value[0][np.roll(difference_greater_one,1)]

        sums_list = []

        for i in range(len(first_pos_indexes)):
            summed_energy = sum(power[first_pos_indexes[i]:last_pos_indexes[i] + 1]) * \
                            timestep / 3600 / 1000

            if summed_energy != 0:
                sums_list.append(summed_energy)

        if sums_list:
            pos_energy_swapsign_norm = np.mean(sums_list) / nom_capacity * 100  # in %
        else:
            pos_energy_swapsign_norm = 'No energy was charged'

        return pos_energy_swapsign_norm

    @property
    def storage_technology_temporal_utilization(self) -> float:
        """
        calculates and returns storage technology temporal utilization in p.u.
        :return: storage technlogy temporal utilization as float
        :return:
        """
        data = self.get_data()
        dc_power = data.dc_power
        t_simulation = len(dc_power)
        t_operation = len(np.where(dc_power != 0)[0])
        return t_operation/t_simulation

    @property
    def average_fulfillment(self) -> float:
        """
        Calculates the average fulfillment factor of the system/battery. How often can the battery/system charge/discharge the desired amount of power.

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            average fulfillment factor
        """
        data: Data = self.get_data()
        return 100 * np.mean(data.storage_fulfillment)

    def close(self) -> None:
        self.__log.close()



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.calendar.no_calendar_degradation import \
    NoCalendarDegradationModel
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.no_cyclic_degradation import \
    NoCyclicDegradationModel
from simses.technology.hydrogen.electrolyzer.degradation.degradation_model import \
    DegradationModel


class NoDegradationModel(DegradationModel):

    def __init__(self):
        super().__init__(NoCyclicDegradationModel(), NoCalendarDegradationModel())

    def calculate_soh(self, state: ElectrolyzerState) -> None:
        pass

    def get_soh_el(self):
        return 1



from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class PowerFollower(OperationStrategy):

    """
    PowerFollower is a basic operation strategy which just forwards a given power profile to the storage system.
    Here, SimSES interprets positive values as charging and negative as discharging of the storage.
    """

    def __init__(self, power_profile: PowerProfile):
        super().__init__(OperationPriority.MEDIUM)
        self.__power_profile: PowerProfile = power_profile
        self.__power: float = 0.0

    def next(self, time: float, system_state: SystemState, power: float = 0) -> float:
        self.__power = self.__power_profile.next(time)
        return 1.0 * self.__power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.load_power = self.__power

    def clear(self) -> None:
        self.__power = 0.0

    def close(self) -> None:
        self.__power_profile.close()



import math
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel


class SanyoNMCCalendarDegradationModel(CalendarDegradationModel):
    """Source: Schmalstieg, J., Käbitz, S., Ecker, M., & Sauer, D. U. (2014).
    A holistic aging model for Li (NiMnCo) O2 based 18650 lithium-ion batteries.
    Journal of Power Sources, 257, 325-334."""
    __capacity_loss = 0
    __resistance_increase = 0

    def __init__(self, cell_type: CellType):
        super().__init__(cell_type)
        self.__time_total_cap = 0
        self.__time_total_res = 0

    def calculate_degradation(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        temp = battery_state.temperature  # cell temperature in K
        voltage = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()  # single cell voltage
        voltage = max(voltage, 3.149)

        alpha_cap = (7.543 * voltage - 23.75)*10**6*math.exp(-6976/temp)
        # implementation of discretization method as in MA Chuanqin Ni
        total_capacity_t1 = 1 - alpha_cap * (self.__time_total_cap / 86400) ** 0.75  # in p.u.
        total_capacity_t2 = 1 - alpha_cap * ((self.__time_total_cap + time_passed) / 86400) ** 0.75  # in p.u.
        capacity_loss = total_capacity_t1 - total_capacity_t2

        self.__time_total_cap += time_passed
        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # Ah

    def calculate_resistance_increase(self, time: float, battery_state: LithiumIonState) -> None:
        time_passed = time - battery_state.time
        temp = battery_state.temperature  # cell temperature in K
        voltage = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale() # single cell voltage
        voltage = max(voltage, 3.097)

        # implementation of discretization method as in MA Chuanqin Ni
        alpha_res = (5.27*voltage - 16.32)*10**5*math.exp(-5986/temp)
        total_resistance_t1 = 1 + alpha_res * (self.__time_total_res / 86400) ** 0.75  # in p.u.
        total_resistance_t2 = 1 + alpha_res * ((self.__time_total_res + time_passed) / 86400) ** 0.75  # in p.u.

        self.__time_total_res += time_passed
        self.__resistance_increase = total_resistance_t2 - total_resistance_t1  # pu

    def get_degradation(self) -> float:
        return self.__capacity_loss

    def get_resistance_increase(self) -> float:
        return self.__resistance_increase

    def reset(self, battery_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        pass



import datetime
from configparser import ConfigParser

import numpy as np
import pandas
import pytest
import pytz

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.demand_charge_reduction import DemandChargeReduction
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig

# fixed input parameters
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState

yearstart = 2020
yearend = 2025
timestep = 900
demand_charge_price = 120
demand_charge_intervall = 900
load_const = 4000
pv_const = 2000


def create_general_config() -> GeneralSimulationConfig:
    simulation_config: ConfigParser = ConfigParser()
    simulation_config.add_section('GENERAL')
    simulation_config.set('GENERAL', 'TIME_STEP', str(timestep))
    return GeneralSimulationConfig(simulation_config)


def create_economic_analysis_config(billing_cycle) -> EconomicAnalysisConfig:
    analysis_config: ConfigParser = ConfigParser()
    analysis_config.add_section('ECONOMIC_ANALYSIS')
    analysis_config.set('ECONOMIC_ANALYSIS', 'DEMAND_CHARGE_BILLING_PERIOD', str(billing_cycle))
    analysis_config.set('ECONOMIC_ANALYSIS', 'DEMAND_CHARGE_AVERAGE_INTERVAL', str(demand_charge_intervall))
    analysis_config.set('ECONOMIC_ANALYSIS', 'DEMAND_CHARGE_PRICE', str(demand_charge_price))
    return EconomicAnalysisConfig(analysis_config)


@pytest.mark.parametrize('billing_cycle, batt_const',
                          [
                              (DemandChargeReduction.BillingPeriod.MONTHLY, -2000),
                              (DemandChargeReduction.BillingPeriod.MONTHLY, 2000),
                              (DemandChargeReduction.BillingPeriod.YEARLY, -2000),
                              (DemandChargeReduction.BillingPeriod.YEARLY, 2000)
                          ]
                         )  # load convention is used for battery in SimSES
def test_demand_charge_reduction(billing_cycle, batt_const):
    """Performs a unit test by comparing the expected result for a generic
    time series with the actual result."""

    # set up test data
    years = yearend - yearstart
    dstart = datetime.datetime(yearstart, 1, 1, tzinfo=pytz.UTC).timestamp()
    dend = datetime.datetime(yearend, 1, 1, tzinfo=pytz.UTC).timestamp()
    n = int((dend - dstart) / timestep)
    time_test = pandas.array([dstart + timestep * i for i in range(n)])
    load_power = pandas.array([load_const] * n)
    pv_power = pandas.array([pv_const] * n)
    battery_power = pandas.array([batt_const] * n)

    # handle fractional years (leap years -> 1 project year + 1 day)
    years_actual = (dend - dstart) / (365 * 24 * 3600)
    years_fractional = years_actual - years

    # calculate expected result
    demand_charge_base = abs(load_const - pv_const) * demand_charge_price / 1000
    demand_charge_battery = abs(load_const - pv_const + batt_const) * demand_charge_price / 1000
    if billing_cycle == DemandChargeReduction.BillingPeriod.MONTHLY:
        demand_charge_base *= 12
        demand_charge_battery *= 12
    expected_result = np.array([demand_charge_base - demand_charge_battery] * years)
    if years_fractional > 0:
        expected_result = np.append(expected_result, (years_fractional * (demand_charge_base - demand_charge_battery)))

    # build configs for testing
    gen_sim_config = create_general_config()
    economic_config = create_economic_analysis_config(billing_cycle)

    # create data for testing
    energy_management_dict = {EnergyManagementState.TIME: time_test, EnergyManagementState.LOAD_POWER: load_power,
                              EnergyManagementState.PV_POWER: pv_power}
    energy_management_data = EnergyManagementData(gen_sim_config, pandas.DataFrame(energy_management_dict))
    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: battery_power}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    # get results
    self_demand_charge_rev_stream = DemandChargeReduction(energy_management_data, system_data, economic_config)
    result1 = self_demand_charge_rev_stream.get_cashflow()

    # Note: in leap years, the demand chargers calculated to project years will be lower
    # In that case, demand charges are spread over 366 days, but each project year only considers 365 days
    # Therefore this tests only checks if results is within +-5 of the exepected result
    expected_upper = expected_result * 1.05
    expected_lower = expected_result * 0.95
    assert list(abs(expected_upper.round(0))) >= list(abs(result1.round(0))) >= list(abs(expected_lower.round(0)))